from pyquil.quil import Program
from pyquil.quilatom import Qubit, Frame
from pyquil.quilbase import DefFrame
from pyquil.api._rewrite_arithmetic import rewrite_arithmetic
from rpcq.messages import (
    ParameterAref,
    ParameterSpec,
    RewriteArithmeticResponse,
)


def test_rewrite_arithmetic_no_params():
    prog = Program("X 0")
    response = rewrite_arithmetic(prog)
    assert response == RewriteArithmeticResponse(quil=Program("X 0").out())


def test_rewrite_arithmetic_simple_mref():
    prog = Program("DECLARE theta REAL", "RZ(theta) 0")
    response = rewrite_arithmetic(prog)
    assert response == RewriteArithmeticResponse(
        original_memory_descriptors={"theta": ParameterSpec(length=1, type="REAL")},
        quil=Program("DECLARE __P1 REAL[1]", "DECLARE theta REAL[1]", "RZ(__P1[0]) 0").out(),
        recalculation_table={ParameterAref(index=0, name="__P1"): "theta[0]/(2*pi)"},
    )


def test_rewrite_arithmetic_duplicate_exprs():
    prog = Program(
        "DECLARE theta REAL",
        "RZ(theta*1.5) 0",
        "RX(theta*1.5) 0",  # this is not a native gate, but it is a protoquil program
    )

    response = rewrite_arithmetic(prog)

    assert response == RewriteArithmeticResponse(
        original_memory_descriptors={"theta": ParameterSpec(length=1, type="REAL")},
        recalculation_table={ParameterAref(index=0, name="__P1"): "theta[0]*1.5/(2*pi)"},
        quil=Program("DECLARE __P1 REAL[1]", "DECLARE theta REAL[1]", "RZ(__P1[0]) 0", "RX(__P1[0]) 0").out(),
    )


def test_rewrite_arithmetic_mixed():
    prog = Program(
        "DECLARE theta REAL",
        "DECLARE beta REAL",
        "RZ(3 * theta) 0",
        "RZ(beta+theta) 0",
    )
    response = rewrite_arithmetic(prog)
    assert response.original_memory_descriptors == {
        "theta": ParameterSpec(length=1, type="REAL"),
        "beta": ParameterSpec(length=1, type="REAL"),
    }
    assert response.recalculation_table == {
        ParameterAref(index=0, name="__P2"): "3*theta[0]/(2*pi)",
        ParameterAref(index=1, name="__P2"): "(beta[0] + theta[0])/(2*pi)",
    }
    assert (
        response.quil
        == Program(
            "DECLARE __P2 REAL[2]",
            "DECLARE theta REAL[1]",
            "DECLARE beta REAL[1]",
            "RZ(__P2[0]) 0",
            "RZ(__P2[1]) 0",
        ).out()
    )


def test_rewrite_arithmetic_set_scale():
    prog = Program(
        "DECLARE theta REAL",
        'SET-SCALE 0 "rf" 1.0',
        'SET-SCALE 0 "rf" theta',
    )

    response = rewrite_arithmetic(prog)

    assert response == RewriteArithmeticResponse(
        original_memory_descriptors={"theta": ParameterSpec(length=1, type="REAL")},
        recalculation_table={ParameterAref(index=0, name="__P1"): "theta[0]/8"},
        quil=Program(
            "DECLARE __P1 REAL[1]",
            "DECLARE theta REAL[1]",
            'SET-SCALE 0 "rf" 1.0',
            'SET-SCALE 0 "rf" __P1[0]',
        ).out(),
    )


def test_rewrite_arithmetic_frequency():
    fdefn0 = DefFrame(
        frame=Frame([Qubit(0)], "rf"),
        center_frequency=10.0,
        sample_rate=20.0,
    )
    fdefn1 = DefFrame(
        frame=Frame([Qubit(1)], "rf"),
        sample_rate=20.0,
    )
    prog = Program(
        fdefn0,
        fdefn1,
        "DECLARE theta REAL",
        'SET-FREQUENCY 0 "rf" theta',
        'SHIFT-FREQUENCY 0 "rf" theta',
        'SET-FREQUENCY 1 "rf" theta',
    )

    response = rewrite_arithmetic(prog)

    assert response == RewriteArithmeticResponse(
        original_memory_descriptors={"theta": ParameterSpec(length=1, type="REAL")},
        recalculation_table={
            ParameterAref(index=0, name="__P1"): "(theta[0] - 10.0)/20.0",
            ParameterAref(index=1, name="__P1"): "theta[0]/20.0",
        },
        quil=Program(
            fdefn0,
            fdefn1,
            "DECLARE __P1 REAL[2]",
            "DECLARE theta REAL[1]",
            'SET-FREQUENCY 0 "rf" __P1[0]',
            'SHIFT-FREQUENCY 0 "rf" __P1[0]',
            'SET-FREQUENCY 1 "rf" __P1[1]',
        ).out(),
    )


def test_rewrite_arithmetic_mixed_mutations():
    fdefn = DefFrame(
        frame=Frame([Qubit(0)], "rf"),
        center_frequency=10.0,
        sample_rate=20.0,
    )
    prog = Program(
        fdefn,
        "DECLARE theta REAL",
        'SET-FREQUENCY 0 "rf" theta',
        'SET-PHASE 0 "rf" theta',
        'SET-SCALE 0 "rf" theta',
    )

    response = rewrite_arithmetic(prog)

    assert response == RewriteArithmeticResponse(
        original_memory_descriptors={"theta": ParameterSpec(length=1, type="REAL")},
        recalculation_table={
            ParameterAref(index=0, name="__P1"): "(theta[0] - 10.0)/20.0",
            ParameterAref(index=1, name="__P1"): "theta[0]/(2*pi)",
            ParameterAref(index=2, name="__P1"): "theta[0]/8",
        },
        quil=Program(
            fdefn,
            "DECLARE __P1 REAL[3]",
            "DECLARE theta REAL[1]",
            'SET-FREQUENCY 0 "rf" __P1[0]',
            'SET-PHASE 0 "rf" __P1[1]',
            'SET-SCALE 0 "rf" __P1[2]',
        ).out(),
    )



import itertools
import random
from math import pi

import numpy as np
import pytest

from pyquil import Program, get_qc
from pyquil.api import WavefunctionSimulator, Client
from pyquil.experiment import (
    ExperimentSetting,
    SIC0,
    SIC1,
    SIC2,
    SIC3,
    TensorProductState,
    Experiment,
    minusY,
    minusZ,
    plusX,
    plusY,
    plusZ,
)
from pyquil.gates import CNOT, CZ, H, I, RX, RY, RZ, X, Y
from pyquil.operator_estimation import (
    _one_q_sic_prep,
    group_experiments,
    measure_observables,
    _ops_bool_to_prog,
    _stats_from_measurements,
    _calibration_program,
)
from pyquil.paulis import sI, sX, sY, sZ, PauliSum
from pyquil.quilbase import Pragma


def test_measure_observables(client: Client):
    expts = [
        ExperimentSetting(TensorProductState(), o1 * o2)
        for o1, o2 in itertools.product([sI(0), sX(0), sY(0), sZ(0)], [sI(1), sX(1), sY(1), sZ(1)])
    ]
    suite = Experiment(expts, program=Program(X(0), CNOT(0, 1)))
    assert len(suite) == 4 * 4
    gsuite = group_experiments(suite)
    assert len(gsuite) == 3 * 3  # can get all the terms with I for free in this case

    qc = get_qc("2q-qvm", client=client)
    for res in measure_observables(qc, gsuite, n_shots=2000):
        if res.setting.out_operator in [sI(), sZ(0), sZ(1), sZ(0) * sZ(1)]:
            assert np.abs(res.expectation) > 0.9
        else:
            assert np.abs(res.expectation) < 0.1


def _random_2q_programs(n_progs=3):
    """Generate random programs that consist of single qubit rotations, a CZ, and single
    qubit rotations.
    """
    r = random.Random(52)

    def RI(qubit, angle):
        # throw away angle so we can randomly choose the identity
        return I(qubit)

    def _random_1q_gate(qubit):
        return r.choice([RI, RX, RY, RZ])(qubit=qubit, angle=r.uniform(0, 2 * pi))

    for _ in range(n_progs):
        prog = Program()
        prog += _random_1q_gate(0)
        prog += _random_1q_gate(1)
        prog += CZ(0, 1)
        prog += _random_1q_gate(0)
        prog += _random_1q_gate(1)
        yield prog


@pytest.mark.slow
def test_measure_observables_many_progs(client: Client):
    expts = [
        ExperimentSetting(TensorProductState(), o1 * o2)
        for o1, o2 in itertools.product([sI(0), sX(0), sY(0), sZ(0)], [sI(1), sX(1), sY(1), sZ(1)])
    ]

    qc = get_qc("2q-qvm", client=client)
    qc.qam.random_seed = 0
    for prog in _random_2q_programs():
        suite = Experiment(expts, program=prog)
        assert len(suite) == 4 * 4
        gsuite = group_experiments(suite)
        assert len(gsuite) == 3 * 3  # can get all the terms with I for free in this case

        wfn = WavefunctionSimulator(client=client)
        wfn_exps = {}
        for expt in expts:
            wfn_exps[expt] = wfn.expectation(gsuite.program, PauliSum([expt.out_operator]))

        for res in measure_observables(qc, gsuite):
            np.testing.assert_allclose(wfn_exps[res.setting], res.expectation, atol=2e-2)


def test_append():
    expts = [
        [
            ExperimentSetting(TensorProductState(), sX(0) * sI(1)),
            ExperimentSetting(TensorProductState(), sI(0) * sX(1)),
        ],
        [
            ExperimentSetting(TensorProductState(), sZ(0) * sI(1)),
            ExperimentSetting(TensorProductState(), sI(0) * sZ(1)),
        ],
    ]
    suite = Experiment(settings=expts, program=Program(X(0), Y(1)))
    suite.append(ExperimentSetting(TensorProductState(), sY(0) * sX(1)))
    assert (len(str(suite))) > 0


def test_no_complex_coeffs(client: Client):
    qc = get_qc("2q-qvm", client=client)
    suite = Experiment([ExperimentSetting(TensorProductState(), 1.0j * sY(0))], program=Program(X(0)))
    with pytest.raises(ValueError):
        list(measure_observables(qc, suite, n_shots=2000))


def test_identity(client: Client):
    qc = get_qc("2q-qvm", client=client)
    suite = Experiment([ExperimentSetting(plusZ(0), 0.123 * sI(0))], program=Program(X(0)))
    result = list(measure_observables(qc, suite))[0]
    assert result.expectation == 0.123


def test_sic_process_tomo(client: Client):
    qc = get_qc("2q-qvm", client=client)
    process = Program(X(0))
    settings = []
    for in_state in [SIC0, SIC1, SIC2, SIC3]:
        for out_op in [sI, sX, sY, sZ]:
            settings += [ExperimentSetting(in_state=in_state(q=0), out_operator=out_op(q=0))]

    experiment = Experiment(settings=settings, program=process)
    results = list(measure_observables(qc, experiment))
    assert len(results) == 4 * 4


def test_measure_observables_symmetrize(client: Client):
    """
    Symmetrization alone should not change the outcome on the QVM
    """
    expts = [
        ExperimentSetting(TensorProductState(), o1 * o2)
        for o1, o2 in itertools.product([sI(0), sX(0), sY(0), sZ(0)], [sI(1), sX(1), sY(1), sZ(1)])
    ]
    suite = Experiment(expts, program=Program(X(0), CNOT(0, 1)))
    assert len(suite) == 4 * 4
    gsuite = group_experiments(suite)
    assert len(gsuite) == 3 * 3  # can get all the terms with I for free in this case

    qc = get_qc("2q-qvm", client=client)
    for res in measure_observables(qc, gsuite, calibrate_readout=None):
        if res.setting.out_operator in [sI(), sZ(0), sZ(1), sZ(0) * sZ(1)]:
            assert np.abs(res.expectation) > 0.9
        else:
            assert np.abs(res.expectation) < 0.1


def test_measure_observables_symmetrize_calibrate(client: Client):
    """
    Symmetrization + calibration should not change the outcome on the QVM
    """
    expts = [
        ExperimentSetting(TensorProductState(), o1 * o2)
        for o1, o2 in itertools.product([sI(0), sX(0), sY(0), sZ(0)], [sI(1), sX(1), sY(1), sZ(1)])
    ]
    suite = Experiment(expts, program=Program(X(0), CNOT(0, 1)))
    assert len(suite) == 4 * 4
    gsuite = group_experiments(suite)
    assert len(gsuite) == 3 * 3  # can get all the terms with I for free in this case

    qc = get_qc("2q-qvm", client=client)
    for res in measure_observables(qc, gsuite):
        if res.setting.out_operator in [sI(), sZ(0), sZ(1), sZ(0) * sZ(1)]:
            assert np.abs(res.expectation) > 0.9
        else:
            assert np.abs(res.expectation) < 0.1


def test_measure_observables_zero_expectation(client: Client):
    """
    Testing case when expectation value of observable should be close to zero
    """
    qc = get_qc("2q-qvm", client=client)
    exptsetting = ExperimentSetting(plusZ(0), sX(0))
    suite = Experiment([exptsetting], program=Program(I(0)))
    result = list(measure_observables(qc, suite))[0]
    np.testing.assert_almost_equal(result.expectation, 0.0, decimal=1)


def test_measure_observables_no_symm_calibr_raises_error(client: Client):
    qc = get_qc("2q-qvm", client=client)
    exptsetting = ExperimentSetting(plusZ(0), sX(0))
    suite = Experiment([exptsetting], program=Program(I(0)), symmetrization=0)
    with pytest.raises(ValueError):
        list(measure_observables(qc, suite, calibrate_readout="plus-eig"))


def test_ops_bool_to_prog():
    qubits = [0, 2, 3]
    ops_strings = list(itertools.product([0, 1], repeat=len(qubits)))
    d_expected = {
        (0, 0, 0): "",
        (0, 0, 1): "X 3\n",
        (0, 1, 0): "X 2\n",
        (0, 1, 1): "X 2\nX 3\n",
        (1, 0, 0): "X 0\n",
        (1, 0, 1): "X 0\nX 3\n",
        (1, 1, 0): "X 0\nX 2\n",
        (1, 1, 1): "X 0\nX 2\nX 3\n",
    }
    for op_str in ops_strings:
        p = _ops_bool_to_prog(op_str, qubits)
        assert str(p) == d_expected[op_str]


def test_stats_from_measurements():
    bs_results = np.array([[0, 1] * 10])
    d_qub_idx = {0: 0, 1: 1}
    setting = ExperimentSetting(TensorProductState(), sZ(0) * sX(1))
    n_shots = 2000

    obs_mean, obs_var = _stats_from_measurements(bs_results, d_qub_idx, setting, n_shots)
    assert obs_mean == -1.0
    assert obs_var == 0.0


def test_measure_observables_uncalibrated_asymmetric_readout(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        runs = 1
    else:
        runs = 100
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p00, p11 = 0.90, 0.80
    p.define_noisy_readout(0, p00=p00, p11=p11)
    p.wrap_in_numshots_loop(2000)
    expt_list = [expt1, expt2, expt3]
    tomo_expt = Experiment(settings=expt_list * runs, program=p, symmetrization=0)
    expected_expectation_z_basis = 2 * p00 - 1

    expect_arr = np.zeros(runs * len(expt_list))

    for idx, res in enumerate(measure_observables(qc, tomo_expt, calibrate_readout=None)):
        expect_arr[idx] = res.expectation

    assert np.isclose(np.mean(expect_arr[::3]), expected_expectation_z_basis, atol=2e-2)
    assert np.isclose(np.mean(expect_arr[1::3]), expected_expectation_z_basis, atol=2e-2)
    assert np.isclose(np.mean(expect_arr[2::3]), expected_expectation_z_basis, atol=2e-2)


def test_measure_observables_uncalibrated_symmetric_readout(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        runs = 1
    else:
        runs = 100
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p00, p11 = 0.90, 0.80
    p.define_noisy_readout(0, p00=p00, p11=p11)
    expt_list = [expt1, expt2, expt3]
    tomo_expt = Experiment(settings=expt_list * runs, program=p)
    expected_symm_error = (p00 + p11) / 2
    expected_expectation_z_basis = expected_symm_error * (1) + (1 - expected_symm_error) * (-1)

    uncalibr_e = np.zeros(runs * len(expt_list))

    for idx, res in enumerate(measure_observables(qc, tomo_expt, n_shots=2000, calibrate_readout=None)):
        uncalibr_e[idx] = res.expectation

    assert np.isclose(np.mean(uncalibr_e[::3]), expected_expectation_z_basis, atol=2e-2)
    assert np.isclose(np.mean(uncalibr_e[1::3]), expected_expectation_z_basis, atol=2e-2)
    assert np.isclose(np.mean(uncalibr_e[2::3]), expected_expectation_z_basis, atol=2e-2)


def test_measure_observables_calibrated_symmetric_readout(client: Client, use_seed: bool):
    # expecting the result +1 for calibrated readout
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p.wrap_in_numshots_loop(2000)
    p.define_noisy_readout(0, p00=0.99, p11=0.80)
    tomo_expt = Experiment(settings=[expt1, expt2, expt3], program=p)

    expectations = []
    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt))
        expectations.append([res.expectation for res in expt_results])
    expectations = np.array(expectations)
    results = np.mean(expectations, axis=0)
    np.testing.assert_allclose(results, 1.0, atol=2e-2)


def test_measure_observables_result_zero_symmetrization_calibration(client: Client, use_seed: bool):
    # expecting expectation value to be 0 with symmetrization/calibration
    qc = get_qc("9q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sZ(0))
    expt2 = ExperimentSetting(TensorProductState(minusZ(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(minusY(0)), sX(0))
    expt_settings = [expt1, expt2, expt3]
    p = Program()
    p00, p11 = 0.99, 0.80
    p.define_noisy_readout(0, p00=p00, p11=p11)
    p.wrap_in_numshots_loop(2000)
    tomo_expt = Experiment(settings=expt_settings, program=p)

    expectations = []
    raw_expectations = []
    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt))
        expectations.append([res.expectation for res in expt_results])
        raw_expectations.append([res.raw_expectation for res in expt_results])
    expectations = np.array(expectations)
    raw_expectations = np.array(raw_expectations)
    results = np.mean(expectations, axis=0)
    raw_results = np.mean(raw_expectations)
    np.testing.assert_allclose(results, 0.0, atol=2e-2)
    np.testing.assert_allclose(raw_results, 0.0, atol=2e-2)


def test_measure_observables_result_zero_no_noisy_readout(client: Client, use_seed: bool):
    # expecting expectation value to be 0 with no symmetrization/calibration
    # and no noisy readout
    qc = get_qc("9q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sZ(0))
    expt2 = ExperimentSetting(TensorProductState(minusZ(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusY(0)), sX(0))
    expt_settings = [expt1, expt2, expt3]
    p = Program()
    tomo_expt = Experiment(settings=expt_settings, program=p, symmetrization=0)

    expectations = []
    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=2000, calibrate_readout=None))
        expectations.append([res.expectation for res in expt_results])
    expectations = np.array(expectations)
    results = np.mean(expectations, axis=0)
    np.testing.assert_allclose(results, 0.0, atol=2e-2)


def test_measure_observables_result_zero_no_symm_calibr(client: Client, use_seed: bool):
    # expecting expectation value to be nonzero with symmetrization/calibration
    qc = get_qc("9q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 3
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sZ(0))
    expt2 = ExperimentSetting(TensorProductState(minusZ(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(minusY(0)), sX(0))
    expt_settings = [expt1, expt2, expt3]
    p = Program()
    p00, p11 = 0.99, 0.80
    p.define_noisy_readout(0, p00=p00, p11=p11)
    tomo_expt = Experiment(settings=expt_settings, program=p, symmetrization=0)

    expectations = []
    expected_result = (p00 * 0.5 + (1 - p11) * 0.5) - ((1 - p00) * 0.5 + p11 * 0.5)
    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=2000, calibrate_readout=None))
        expectations.append([res.expectation for res in expt_results])
    expectations = np.array(expectations)
    results = np.mean(expectations, axis=0)
    np.testing.assert_allclose(results, expected_result, atol=2e-2)


def test_measure_observables_2q_readout_error_one_measured(client: Client, use_seed: bool):
    # 2q readout errors, but only 1 qubit measured
    qc = get_qc("9q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 3
        np.random.seed(0)
        runs = 1
    else:
        runs = 100
    qubs = [0, 1]
    expt = ExperimentSetting(TensorProductState(plusZ(qubs[0]) * plusZ(qubs[1])), sZ(qubs[0]))
    p = Program()
    p.define_noisy_readout(0, 0.999, 0.85)
    p.define_noisy_readout(1, 0.999, 0.75)
    tomo_experiment = Experiment(settings=[expt] * runs, program=p)

    raw_e = np.zeros(runs)
    obs_e = np.zeros(runs)
    cal_e = np.zeros(runs)

    for idx, res in enumerate(measure_observables(qc, tomo_experiment, n_shots=5000)):
        raw_e[idx] = res.raw_expectation
        obs_e[idx] = res.expectation
        cal_e[idx] = res.calibration_expectation

    assert np.isclose(np.mean(raw_e), 0.849, atol=2e-2)
    assert np.isclose(np.mean(obs_e), 1.0, atol=2e-2)
    assert np.isclose(np.mean(cal_e), 0.849, atol=2e-2)


def test_measure_observables_inherit_noise_errors(client: Client):
    qc = get_qc("3q-qvm", client=client)
    # specify simplest experiments
    expt1 = ExperimentSetting(TensorProductState(), sZ(0))
    expt2 = ExperimentSetting(TensorProductState(), sZ(1))
    expt3 = ExperimentSetting(TensorProductState(), sZ(2))
    # specify a Program with multiple sources of noise
    p = Program(X(0), Y(1), H(2))
    # defining several bit-flip channels
    kraus_ops_X = [
        np.sqrt(1 - 0.3) * np.array([[1, 0], [0, 1]]),
        np.sqrt(0.3) * np.array([[0, 1], [1, 0]]),
    ]
    kraus_ops_Y = [
        np.sqrt(1 - 0.2) * np.array([[1, 0], [0, 1]]),
        np.sqrt(0.2) * np.array([[0, 1], [1, 0]]),
    ]
    kraus_ops_H = [
        np.sqrt(1 - 0.1) * np.array([[1, 0], [0, 1]]),
        np.sqrt(0.1) * np.array([[0, 1], [1, 0]]),
    ]
    # replacing all the gates with bit-flip channels
    p.define_noisy_gate("X", [0], kraus_ops_X)
    p.define_noisy_gate("Y", [1], kraus_ops_Y)
    p.define_noisy_gate("H", [2], kraus_ops_H)
    # defining readout errors
    p.define_noisy_readout(0, 0.99, 0.80)
    p.define_noisy_readout(1, 0.95, 0.85)
    p.define_noisy_readout(2, 0.97, 0.78)

    tomo_expt = Experiment(settings=[expt1, expt2, expt3], program=p)

    calibr_prog1 = _calibration_program(qc, tomo_expt, expt1)
    calibr_prog2 = _calibration_program(qc, tomo_expt, expt2)
    calibr_prog3 = _calibration_program(qc, tomo_expt, expt3)
    expected_prog = """PRAGMA READOUT-POVM 0 "(0.99 0.19999999999999996 0.010000000000000009 0.8)"
PRAGMA READOUT-POVM 1 "(0.95 0.15000000000000002 0.050000000000000044 0.85)"
PRAGMA READOUT-POVM 2 "(0.97 0.21999999999999997 0.030000000000000027 0.78)"
PRAGMA ADD-KRAUS X 0 "(0.8366600265340756 0.0 0.0 0.8366600265340756)"
PRAGMA ADD-KRAUS X 0 "(0.0 0.5477225575051661 0.5477225575051661 0.0)"
PRAGMA ADD-KRAUS Y 1 "(0.8944271909999159 0.0 0.0 0.8944271909999159)"
PRAGMA ADD-KRAUS Y 1 "(0.0 0.4472135954999579 0.4472135954999579 0.0)"
PRAGMA ADD-KRAUS H 2 "(0.9486832980505138 0.0 0.0 0.9486832980505138)"
PRAGMA ADD-KRAUS H 2 "(0.0 0.31622776601683794 0.31622776601683794 0.0)"
"""
    assert calibr_prog1.out() == Program(expected_prog).out()
    assert calibr_prog2.out() == Program(expected_prog).out()
    assert calibr_prog3.out() == Program(expected_prog).out()


def test_expectations_sic0(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(SIC0(0), sX(0))
    expt2 = ExperimentSetting(SIC0(0), sY(0))
    expt3 = ExperimentSetting(SIC0(0), sZ(0))
    tomo_expt = Experiment(settings=[expt1, expt2, expt3], program=Program())

    results_unavged = []
    for _ in range(num_simulations):
        measured_results = []
        for res in measure_observables(qc, tomo_expt, n_shots=2000):
            measured_results.append(res.expectation)
        results_unavged.append(measured_results)

    results_unavged = np.array(results_unavged)
    results = np.mean(results_unavged, axis=0)
    expected_results = np.array([0, 0, 1])
    np.testing.assert_allclose(results, expected_results, atol=2e-2)


def test_expectations_sic1(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(SIC1(0), sX(0))
    expt2 = ExperimentSetting(SIC1(0), sY(0))
    expt3 = ExperimentSetting(SIC1(0), sZ(0))
    tomo_expt = Experiment(settings=[expt1, expt2, expt3], program=Program())

    results_unavged = []
    for _ in range(num_simulations):
        measured_results = []
        for res in measure_observables(qc, tomo_expt, n_shots=2000):
            measured_results.append(res.expectation)
        results_unavged.append(measured_results)

    results_unavged = np.array(results_unavged)
    results = np.mean(results_unavged, axis=0)
    expected_results = np.array([2 * np.sqrt(2) / 3, 0, -1 / 3])
    np.testing.assert_allclose(results, expected_results, atol=2e-2)


def test_expectations_sic2(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(SIC2(0), sX(0))
    expt2 = ExperimentSetting(SIC2(0), sY(0))
    expt3 = ExperimentSetting(SIC2(0), sZ(0))
    tomo_expt = Experiment(settings=[expt1, expt2, expt3], program=Program())

    results_unavged = []
    for _ in range(num_simulations):
        measured_results = []
        for res in measure_observables(qc, tomo_expt, n_shots=2000):
            measured_results.append(res.expectation)
        results_unavged.append(measured_results)

    results_unavged = np.array(results_unavged)
    results = np.mean(results_unavged, axis=0)
    expected_results = np.array(
        [
            (2 * np.sqrt(2) / 3) * np.cos(2 * np.pi / 3),
            -(2 * np.sqrt(2) / 3) * np.sin(2 * np.pi / 3),
            -1 / 3,
        ]
    )
    np.testing.assert_allclose(results, expected_results, atol=2e-2)


def test_expectations_sic3(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt1 = ExperimentSetting(SIC3(0), sX(0))
    expt2 = ExperimentSetting(SIC3(0), sY(0))
    expt3 = ExperimentSetting(SIC3(0), sZ(0))
    tomo_expt = Experiment(settings=[expt1, expt2, expt3], program=Program())

    results_unavged = []
    for _ in range(num_simulations):
        measured_results = []
        for res in measure_observables(qc, tomo_expt, n_shots=2000):
            measured_results.append(res.expectation)
        results_unavged.append(measured_results)

    results_unavged = np.array(results_unavged)
    results = np.mean(results_unavged, axis=0)
    expected_results = np.array(
        [
            (2 * np.sqrt(2) / 3) * np.cos(2 * np.pi / 3),
            (2 * np.sqrt(2) / 3) * np.sin(2 * np.pi / 3),
            -1 / 3,
        ]
    )
    np.testing.assert_allclose(results, expected_results, atol=2e-2)


def test_sic_conditions(client: Client):
    """
    Test that the SIC states indeed yield SIC-POVMs
    """
    wfn_sim = WavefunctionSimulator(client=client)

    # condition (i) -- sum of all projectors equal identity times dimensionality
    result = np.zeros((2, 2))

    for i in range(4):
        if i == 0:
            amps = np.array([1, 0])
        else:
            sic = _one_q_sic_prep(i, 0)
            wfn = wfn_sim.wavefunction(sic)
            amps = wfn.amplitudes
        proj = np.outer(amps, amps.conj())
        result = np.add(result, proj)
    np.testing.assert_allclose(result / 2, np.eye(2), atol=2e-2)

    # condition (ii) -- tr(proj_a . proj_b) = 1 / 3, for a != b
    for comb in itertools.combinations([0, 1, 2, 3], 2):
        if comb[0] == 0:
            sic_a = Program(I(0))
        else:
            sic_a = _one_q_sic_prep(comb[0], 0)
        sic_b = _one_q_sic_prep(comb[1], 0)

        wfn_a = wfn_sim.wavefunction(sic_a)
        wfn_b = wfn_sim.wavefunction(sic_b)

        amps_a = wfn_a.amplitudes
        amps_b = wfn_b.amplitudes

        proj_a = np.outer(amps_a, amps_a.conj())
        proj_b = np.outer(amps_b, amps_b.conj())

        assert np.isclose(np.trace(proj_a.dot(proj_b)), 1 / 3)


def test_measure_observables_grouped_expts(client: Client, use_seed: bool):
    qc = get_qc("3q-qvm", client=client)

    if use_seed:
        num_simulations = 1
        qc.qam.random_seed = 4
    else:
        num_simulations = 100

    # this more explicitly uses the list-of-lists-of-ExperimentSettings in TomographyExperiment
    # create experiments in different groups
    expt1_group1 = ExperimentSetting(SIC1(0) * plusX(1), sZ(0) * sX(1))
    expt2_group1 = ExperimentSetting(plusX(1) * minusY(2), sX(1) * sY(2))
    expts_group1 = [expt1_group1, expt2_group1]

    expt1_group2 = ExperimentSetting(plusX(0) * SIC0(1), sX(0) * sZ(1))
    expt2_group2 = ExperimentSetting(SIC0(1) * minusY(2), sZ(1) * sY(2))
    expt3_group2 = ExperimentSetting(plusX(0) * minusY(2), sX(0) * sY(2))
    expts_group2 = [expt1_group2, expt2_group2, expt3_group2]
    # create a list-of-lists-of-ExperimentSettings
    expt_settings = [expts_group1, expts_group2]
    # and use this to create a TomographyExperiment suite
    tomo_expt = Experiment(settings=expt_settings, program=Program())

    results_unavged = []
    for _ in range(num_simulations):
        measured_results = []
        for res in measure_observables(qc, tomo_expt, n_shots=2000):
            measured_results.append(res.expectation)
        results_unavged.append(measured_results)

    results_unavged = np.array(results_unavged)
    results = np.mean(results_unavged, axis=0)
    expected_results = np.array([-1 / 3, -1, 1, -1, -1])
    np.testing.assert_allclose(results, expected_results, atol=2e-2)


def _point_channel_fidelity_estimate(v, dim=2):
    """:param v: array of expectation values
    :param dim: dimensionality of the Hilbert space"""
    return (1.0 + np.sum(v) + dim) / (dim * (dim + 1))


def test_bit_flip_channel_fidelity(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        np.random.seed(0)
        qc.qam.random_seed = 0
        num_expts = 1
    else:
        num_expts = 100

    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare noisy bit-flip channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # the bit flip channel is composed of two Kraus operations --
    # applying the X gate with probability `prob`, and applying the identity gate
    # with probability `1 - prob`
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[0, 1], [1, 0]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)

    # prepare Experiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = 1 - (2 / 3) * prob
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_dephasing_channel_fidelity(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare noisy dephasing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for the dephasing channel
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = 1 - (2 / 3) * prob
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_depolarizing_channel_fidelity(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare noisy depolarizing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for the depolarizing channel
    kraus_ops = [
        np.sqrt(3 * prob + 1) / 2 * np.array([[1, 0], [0, 1]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, 1], [1, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, -1j], [1j, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = (1 + prob) / 2
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_unitary_channel_fidelity(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare unitary channel as an RY rotation program for some random angle
    theta = np.random.uniform(0.0, 2 * np.pi)
    # unitary (RY) channel
    p = Program(RY(theta, 0))
    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = (1 / 6) * ((2 * np.cos(theta / 2)) ** 2 + 2)
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_bit_flip_channel_fidelity_readout_error(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare noisy bit-flip channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # the bit flip channel is composed of two Kraus operations --
    # applying the X gate with probability `prob`, and applying the identity gate
    # with probability `1 - prob`
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[0, 1], [1, 0]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)
    # add some readout error
    p.define_noisy_readout(0, 0.95, 0.82)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = 1 - (2 / 3) * prob
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_dephasing_channel_fidelity_readout_error(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare noisy dephasing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for the dephasing channel
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)
    # add some readout error
    p.define_noisy_readout(0, 0.95, 0.82)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = 1 - (2 / 3) * prob
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_depolarizing_channel_fidelity_readout_error(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare noisy depolarizing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for the depolarizing channel
    kraus_ops = [
        np.sqrt(3 * prob + 1) / 2 * np.array([[1, 0], [0, 1]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, 1], [1, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, -1j], [1j, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)
    # add some readout error
    p.define_noisy_readout(0, 0.95, 0.82)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = (1 + prob) / 2
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_unitary_channel_fidelity_readout_error(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    """
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings
    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    expt_list = [expt1, expt2, expt3]

    # prepare unitary channel as an RY rotation program for some random angle
    theta = np.random.uniform(0.0, 2 * np.pi)
    # unitary (RY) channel
    p = Program(RY(theta, 0))
    # add some readout error
    p.define_noisy_readout(0, 0.95, 0.82)
    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results)
    # how close is this channel to the identity operator
    expected_fidelity = (1 / 6) * ((2 * np.cos(theta / 2)) ** 2 + 2)
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_2q_unitary_channel_fidelity_readout_error(client: Client, use_seed: bool):
    """
    We use Eqn (5) of https://arxiv.org/abs/quant-ph/0701138 to compare the fidelity
    This tests if our dimensionality factors are correct, even in the presence
    of readout errors
    """
    qc = get_qc("2q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment settings

    expt1 = ExperimentSetting(TensorProductState(plusX(0)), sX(0))
    expt2 = ExperimentSetting(TensorProductState(plusY(0)), sY(0))
    expt3 = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))

    expt4 = ExperimentSetting(TensorProductState(plusX(1)), sX(1))
    expt5 = ExperimentSetting(TensorProductState(plusX(0) * plusX(1)), sX(0) * sX(1))
    expt6 = ExperimentSetting(TensorProductState(plusY(0) * plusX(1)), sY(0) * sX(1))
    expt7 = ExperimentSetting(TensorProductState(plusZ(0) * plusX(1)), sZ(0) * sX(1))

    expt8 = ExperimentSetting(TensorProductState(plusY(1)), sY(1))
    expt9 = ExperimentSetting(TensorProductState(plusX(0) * plusY(1)), sX(0) * sY(1))
    expt10 = ExperimentSetting(TensorProductState(plusY(0) * plusY(1)), sY(0) * sY(1))
    expt11 = ExperimentSetting(TensorProductState(plusZ(0) * plusY(1)), sZ(0) * sY(1))

    expt12 = ExperimentSetting(TensorProductState(plusZ(1)), sZ(1))
    expt13 = ExperimentSetting(TensorProductState(plusX(0) * plusZ(1)), sX(0) * sZ(1))
    expt14 = ExperimentSetting(TensorProductState(plusY(0) * plusZ(1)), sY(0) * sZ(1))
    expt15 = ExperimentSetting(TensorProductState(plusZ(0) * plusZ(1)), sZ(0) * sZ(1))

    expt_list = [
        expt1,
        expt2,
        expt3,
        expt4,
        expt5,
        expt6,
        expt7,
        expt8,
        expt9,
        expt10,
        expt11,
        expt12,
        expt13,
        expt14,
        expt15,
    ]

    # prepare unitary channel as an RY rotation program for some random angle
    theta1, theta2 = np.random.uniform(0.0, 2 * np.pi, size=2)
    # unitary (RY) channel
    p = Program(RY(theta1, 0), RY(theta2, 1))
    # add some readout error
    p.define_noisy_readout(0, 0.95, 0.82)
    p.define_noisy_readout(1, 0.99, 0.73)
    # prepare TomographyExperiment
    process_exp = Experiment(settings=expt_list, program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=5000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_channel_fidelity_estimate(results, dim=4)
    # how close is this channel to the identity operator
    expected_fidelity = (1 / 5) * ((2 * np.cos(theta1 / 2) * np.cos(theta2 / 2)) ** 2 + 1)
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_measure_1q_observable_raw_expectation(client: Client, use_seed: bool):
    # testing that we get correct raw expectation in terms of readout errors
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    expt = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p00, p11 = 0.99, 0.80
    p.define_noisy_readout(0, p00=p00, p11=p11)
    tomo_expt = Experiment(settings=[expt], program=p)

    raw_expectations = []
    for _ in range(num_expts):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=2000))
        raw_expectations.append([res.raw_expectation for res in expt_results])
    raw_expectations = np.array(raw_expectations)
    result = np.mean(raw_expectations, axis=0)

    # calculate expected raw_expectation
    eps_not = (p00 + p11) / 2
    eps = 1 - eps_not
    expected_result = 1 - 2 * eps
    np.testing.assert_allclose(result, expected_result, atol=2e-2)


def test_measure_1q_observable_raw_variance(client: Client, use_seed: bool):
    # testing that we get correct raw std_err in terms of readout errors
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    expt = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p00, p11 = 0.99, 0.80
    p.define_noisy_readout(0, p00=p00, p11=p11)
    tomo_expt = Experiment(settings=[expt], program=p)

    num_shots = 2000

    raw_std_errs = []
    for _ in range(num_expts):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=num_shots))
        raw_std_errs.append([res.raw_std_err for res in expt_results])
    raw_std_errs = np.array(raw_std_errs)
    result = np.mean(raw_std_errs, axis=0)

    # calculate expected raw_expectation
    eps_not = (p00 + p11) / 2
    eps = 1 - eps_not
    expected_result = np.sqrt((1 - (1 - 2 * eps) ** 2) / num_shots)
    np.testing.assert_allclose(result, expected_result, atol=2e-2)


def test_measure_1q_observable_calibration_expectation(client: Client, use_seed: bool):
    # testing that we get correct calibration expectation in terms of readout errors
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    expt = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p00, p11 = 0.93, 0.77
    p.define_noisy_readout(0, p00=p00, p11=p11)
    tomo_expt = Experiment(settings=[expt], program=p)

    calibration_expectations = []
    for _ in range(num_expts):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=2000))
        calibration_expectations.append([res.calibration_expectation for res in expt_results])
    calibration_expectations = np.array(calibration_expectations)
    result = np.mean(calibration_expectations, axis=0)

    # calculate expected raw_expectation
    eps_not = (p00 + p11) / 2
    eps = 1 - eps_not
    expected_result = 1 - 2 * eps
    np.testing.assert_allclose(result, expected_result, atol=2e-2)


def test_measure_1q_observable_calibration_variance(client: Client, use_seed: bool):
    # testing that we get correct calibration std_err in terms of readout errors
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    expt = ExperimentSetting(TensorProductState(plusZ(0)), sZ(0))
    p = Program()
    p00, p11 = 0.93, 0.77
    p.define_noisy_readout(0, p00=p00, p11=p11)
    tomo_expt = Experiment(settings=[expt], program=p)

    num_shots = 2000

    raw_std_errs = []
    for _ in range(num_expts):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=num_shots))
        raw_std_errs.append([res.raw_std_err for res in expt_results])
    raw_std_errs = np.array(raw_std_errs)
    result = np.mean(raw_std_errs, axis=0)

    # calculate expected raw_expectation
    eps_not = (p00 + p11) / 2
    eps = 1 - eps_not
    expected_result = np.sqrt((1 - (1 - 2 * eps) ** 2) / num_shots)
    np.testing.assert_allclose(result, expected_result, atol=2e-2)


def test_uncalibrated_asymmetric_readout_nontrivial_1q_state(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        runs = 1
    else:
        runs = 100
    expt = ExperimentSetting(TensorProductState(), sZ(0))
    # pick some random value for RX rotation
    theta = np.random.uniform(0.0, 2 * np.pi)
    p = Program(RX(theta, 0))
    # pick some random (but sufficiently large) asymmetric readout errors
    p00, p11 = np.random.uniform(0.7, 0.99, size=2)
    p.define_noisy_readout(0, p00=p00, p11=p11)
    expt_list = [expt]
    tomo_expt = Experiment(settings=expt_list * runs, program=p, symmetrization=0)
    # calculate expected expectation value
    amp_sqr0 = (np.cos(theta / 2)) ** 2
    amp_sqr1 = (np.sin(theta / 2)) ** 2
    expected_expectation = (p00 * amp_sqr0 + (1 - p11) * amp_sqr1) - ((1 - p00) * amp_sqr0 + p11 * amp_sqr1)

    expect_arr = np.zeros(runs * len(expt_list))

    for idx, res in enumerate(measure_observables(qc, tomo_expt, n_shots=2000, calibrate_readout=None)):
        expect_arr[idx] = res.expectation

    assert np.isclose(np.mean(expect_arr), expected_expectation, atol=2e-2)


def test_uncalibrated_symmetric_readout_nontrivial_1q_state(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        runs = 1
    else:
        runs = 100
    expt = ExperimentSetting(TensorProductState(), sZ(0))
    # pick some random value for RX rotation
    theta = np.random.uniform(0.0, 2 * np.pi)
    p = Program(RX(theta, 0))
    # pick some random (but sufficiently large) asymmetric readout errors
    p00, p11 = np.random.uniform(0.7, 0.99, size=2)
    p.define_noisy_readout(0, p00=p00, p11=p11)
    expt_list = [expt]
    tomo_expt = Experiment(settings=expt_list * runs, program=p, symmetrization=-1)
    # calculate expected expectation value
    amp_sqr0 = (np.cos(theta / 2)) ** 2
    amp_sqr1 = (np.sin(theta / 2)) ** 2
    symm_prob = (p00 + p11) / 2
    expected_expectation = (symm_prob * amp_sqr0 + (1 - symm_prob) * amp_sqr1) - (
        (1 - symm_prob) * amp_sqr0 + symm_prob * amp_sqr1
    )

    expect_arr = np.zeros(runs * len(expt_list))

    for idx, res in enumerate(measure_observables(qc, tomo_expt, n_shots=2000, calibrate_readout=None)):
        expect_arr[idx] = res.expectation

    assert np.isclose(np.mean(expect_arr), expected_expectation, atol=2e-2)


def test_calibrated_symmetric_readout_nontrivial_1q_state(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        runs = 1
    else:
        runs = 100
    expt = ExperimentSetting(TensorProductState(), sZ(0))
    # pick some random value for RX rotation
    theta = np.random.uniform(0.0, 2 * np.pi)
    p = Program(RX(theta, 0))
    # pick some random (but sufficiently large) asymmetric readout errors
    p00, p11 = np.random.uniform(0.7, 0.99, size=2)
    p.define_noisy_readout(0, p00=p00, p11=p11)
    expt_list = [expt]
    tomo_expt = Experiment(settings=expt_list * runs, program=p, symmetrization=-1)
    # calculate expected expectation value
    amp_sqr0 = (np.cos(theta / 2)) ** 2
    amp_sqr1 = (np.sin(theta / 2)) ** 2
    expected_expectation = amp_sqr0 - amp_sqr1

    expect_arr = np.zeros(runs * len(expt_list))

    for idx, res in enumerate(measure_observables(qc, tomo_expt, n_shots=2000, calibrate_readout="plus-eig")):
        expect_arr[idx] = res.expectation

    assert np.isclose(np.mean(expect_arr), expected_expectation, atol=2e-2)


def test_measure_2q_observable_raw_statistics(client: Client, use_seed: bool):
    """Testing that we get correct exhaustively symmetrized statistics
        in terms of readout errors.
    Note: this only tests for exhaustive symmetrization in the presence
        of uncorrelated errors
    """
    qc = get_qc("2q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt = ExperimentSetting(TensorProductState(), sZ(0) * sZ(1))
    p = Program()
    p00, p11 = 0.99, 0.80
    q00, q11 = 0.93, 0.76
    p.define_noisy_readout(0, p00=p00, p11=p11)
    p.define_noisy_readout(1, p00=q00, p11=q11)
    tomo_expt = Experiment(settings=[expt], program=p)

    num_shots = 5000

    raw_expectations = []
    raw_std_errs = []

    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=num_shots))
        raw_expectations.append([res.raw_expectation for res in expt_results])
        raw_std_errs.append([res.raw_std_err for res in expt_results])

    raw_expectations = np.array(raw_expectations)
    raw_std_errs = np.array(raw_std_errs)
    result_expectation = np.mean(raw_expectations, axis=0)
    result_std_err = np.mean(raw_std_errs, axis=0)

    # calculate relevant conditional probabilities, given |00> state
    # notation used: pijmn means p(ij|mn)
    p0000 = (p00 + p11) * (q00 + q11) / 4
    p0100 = (p00 + p11) * (2 - q00 - q11) / 4
    p1000 = (q00 + q11) * (2 - p00 - p11) / 4
    p1100 = (2 - p00 - p11) * (2 - q00 - q11) / 4
    # calculate expectation value of Z^{\otimes 2}
    z_expectation = (p0000 + p1100) - (p0100 + p1000)
    # calculate standard deviation of the mean
    simulated_std_err = np.sqrt((1 - z_expectation ** 2) / num_shots)
    # compare against simulated results
    np.testing.assert_allclose(result_expectation, z_expectation, atol=2e-2)
    np.testing.assert_allclose(result_std_err, simulated_std_err, atol=2e-2)


def test_raw_statistics_2q_nontrivial_nonentangled_state(client: Client, use_seed: bool):
    """Testing that we get correct exhaustively symmetrized statistics
        in terms of readout errors, even for non-trivial 2q nonentangled states
    Note: this only tests for exhaustive symmetrization in the presence
        of uncorrelated errors
    """
    qc = get_qc("2q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt = ExperimentSetting(TensorProductState(), sZ(0) * sZ(1))
    theta1, theta2 = np.random.uniform(0.0, 2 * np.pi, size=2)
    p = Program(RX(theta1, 0), RX(theta2, 1))
    p00, p11, q00, q11 = np.random.uniform(0.70, 0.99, size=4)
    p.define_noisy_readout(0, p00=p00, p11=p11)
    p.define_noisy_readout(1, p00=q00, p11=q11)
    tomo_expt = Experiment(settings=[expt], program=p)

    num_shots = 5000

    raw_expectations = []
    raw_std_errs = []

    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=num_shots))
        raw_expectations.append([res.raw_expectation for res in expt_results])
        raw_std_errs.append([res.raw_std_err for res in expt_results])
    raw_expectations = np.array(raw_expectations)
    raw_std_errs = np.array(raw_std_errs)
    result_expectation = np.mean(raw_expectations, axis=0)
    result_std_err = np.mean(raw_std_errs, axis=0)

    # calculate relevant conditional probabilities, given |00> state
    # notation used: pijmn means p(ij|mn)
    p0000 = (p00 + p11) * (q00 + q11) / 4
    p0100 = (p00 + p11) * (2 - q00 - q11) / 4
    p1000 = (q00 + q11) * (2 - p00 - p11) / 4
    p1100 = (2 - p00 - p11) * (2 - q00 - q11) / 4
    # calculate relevant conditional probabilities, given |01> state
    p0001 = p0100
    p0101 = p0000
    p1001 = (2 - p00 - p11) * (2 - q00 - q11) / 4
    p1101 = (2 - p00 - p11) * (q00 + q11) / 4
    # calculate relevant conditional probabilities, given |10> state
    p0010 = p1000
    p0110 = p1001
    p1010 = p0000
    p1110 = (p00 + p11) * (2 - q00 - q11) / 4
    # calculate relevant conditional probabilities, given |11> state
    p0011 = p1100
    p0111 = p1101
    p1011 = p1110
    p1111 = p0000
    # calculate amplitudes squared of pure state
    alph00 = (np.cos(theta1 / 2) * np.cos(theta2 / 2)) ** 2
    alph01 = (np.cos(theta1 / 2) * np.sin(theta2 / 2)) ** 2
    alph10 = (np.sin(theta1 / 2) * np.cos(theta2 / 2)) ** 2
    alph11 = (np.sin(theta1 / 2) * np.sin(theta2 / 2)) ** 2
    # calculate probabilities of various bitstrings
    pr00 = p0000 * alph00 + p0001 * alph01 + p0010 * alph10 + p0011 * alph11
    pr01 = p0100 * alph00 + p0101 * alph01 + p0110 * alph10 + p0111 * alph11
    pr10 = p1000 * alph00 + p1001 * alph01 + p1010 * alph10 + p1011 * alph11
    pr11 = p1100 * alph00 + p1101 * alph01 + p1110 * alph10 + p1111 * alph11
    # calculate Z^{\otimes 2} expectation, and error of the mean
    z_expectation = (pr00 + pr11) - (pr01 + pr10)
    simulated_std_err = np.sqrt((1 - z_expectation ** 2) / num_shots)
    # compare against simulated results
    np.testing.assert_allclose(result_expectation, z_expectation, atol=2e-2)
    np.testing.assert_allclose(result_std_err, simulated_std_err, atol=2e-2)


def test_raw_statistics_2q_nontrivial_entangled_state(client: Client, use_seed: bool):
    """Testing that we get correct exhaustively symmetrized statistics
        in terms of readout errors, even for non-trivial 2q entangled states.
    Note: this only tests for exhaustive symmetrization in the presence
        of uncorrelated errors
    """
    qc = get_qc("2q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_simulations = 1
    else:
        num_simulations = 100
    expt = ExperimentSetting(TensorProductState(), sZ(0) * sZ(1))
    theta = np.random.uniform(0.0, 2 * np.pi)
    p = Program(RX(theta, 0), CNOT(0, 1))
    p00, p11, q00, q11 = np.random.uniform(0.70, 0.99, size=4)
    p.define_noisy_readout(0, p00=p00, p11=p11)
    p.define_noisy_readout(1, p00=q00, p11=q11)
    tomo_expt = Experiment(settings=[expt], program=p)

    num_shots = 5000

    raw_expectations = []
    raw_std_errs = []

    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=num_shots))
        raw_expectations.append([res.raw_expectation for res in expt_results])
        raw_std_errs.append([res.raw_std_err for res in expt_results])
    raw_expectations = np.array(raw_expectations)
    raw_std_errs = np.array(raw_std_errs)
    result_expectation = np.mean(raw_expectations, axis=0)
    result_std_err = np.mean(raw_std_errs, axis=0)

    # calculate relevant conditional probabilities, given |00> state
    # notation used: pijmn means p(ij|mn)
    p0000 = (p00 + p11) * (q00 + q11) / 4
    p0100 = (p00 + p11) * (2 - q00 - q11) / 4
    p1000 = (q00 + q11) * (2 - p00 - p11) / 4
    p1100 = (2 - p00 - p11) * (2 - q00 - q11) / 4
    # calculate relevant conditional probabilities, given |11> state
    p0011 = p1100
    p0111 = (2 - p00 - p11) * (q00 + q11) / 4
    p1011 = (p00 + p11) * (2 - q00 - q11) / 4
    p1111 = p0000
    # calculate amplitudes squared of pure state
    alph00 = (np.cos(theta / 2)) ** 2
    alph11 = (np.sin(theta / 2)) ** 2
    # calculate probabilities of various bitstrings
    pr00 = p0000 * alph00 + p0011 * alph11
    pr01 = p0100 * alph00 + p0111 * alph11
    pr10 = p1000 * alph00 + p1011 * alph11
    pr11 = p1100 * alph00 + p1111 * alph11
    # calculate Z^{\otimes 2} expectation, and error of the mean
    z_expectation = (pr00 + pr11) - (pr01 + pr10)
    simulated_std_err = np.sqrt((1 - z_expectation ** 2) / num_shots)
    # compare against simulated results
    np.testing.assert_allclose(result_expectation, z_expectation, atol=2e-2)
    np.testing.assert_allclose(result_std_err, simulated_std_err, atol=2e-2)


@pytest.mark.flaky(reruns=1)
def test_corrected_statistics_2q_nontrivial_nonentangled_state(client: Client, use_seed: bool):
    """Testing that we can successfully correct for observed statistics
        in the presence of readout errors, even for 2q nontrivial but
        nonentangled states.
    Note: this only tests for exhaustive symmetrization in the presence
        of uncorrelated errors
    """
    qc = get_qc("2q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(13)
        num_simulations = 1
    else:
        num_simulations = 100

    expt = ExperimentSetting(TensorProductState(), sZ(0) * sZ(1))
    theta1, theta2 = np.random.uniform(0.0, 2 * np.pi, size=2)
    p = Program(RX(theta1, 0), RX(theta2, 1))
    p00, p11, q00, q11 = np.random.uniform(0.70, 0.99, size=4)
    p.define_noisy_readout(0, p00=p00, p11=p11)
    p.define_noisy_readout(1, p00=q00, p11=q11)
    tomo_expt = Experiment(settings=[expt], program=p)

    num_shots = 5000

    expectations = []
    std_errs = []

    for _ in range(num_simulations):
        expt_results = list(measure_observables(qc, tomo_expt, n_shots=num_shots))
        expectations.append([res.expectation for res in expt_results])
        std_errs.append([res.std_err for res in expt_results])
    expectations = np.array(expectations)
    std_errs = np.array(std_errs)
    result_expectation = np.mean(expectations, axis=0)
    result_std_err = np.mean(std_errs, axis=0)

    # calculate amplitudes squared of pure state
    alph00 = (np.cos(theta1 / 2) * np.cos(theta2 / 2)) ** 2
    alph01 = (np.cos(theta1 / 2) * np.sin(theta2 / 2)) ** 2
    alph10 = (np.sin(theta1 / 2) * np.cos(theta2 / 2)) ** 2
    alph11 = (np.sin(theta1 / 2) * np.sin(theta2 / 2)) ** 2
    # calculate Z^{\otimes 2} expectation, and error of the mean
    expected_expectation = (alph00 + alph11) - (alph01 + alph10)
    expected_std_err = np.sqrt(np.var(expectations))
    # compare against simulated results
    np.testing.assert_allclose(result_expectation, expected_expectation, atol=2e-2)
    np.testing.assert_allclose(result_std_err, expected_std_err, atol=2e-2)


def _point_state_fidelity_estimate(v, dim=2):
    """:param v: array of expectation values
    :param dim: dimensionality of the Hilbert space"""
    return (1.0 + np.sum(v)) / dim


def test_bit_flip_state_fidelity(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # prepare noisy bit-flip channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # the bit flip channel is composed of two Kraus operations --
    # applying the X gate with probability `prob`, and applying the identity gate
    # with probability `1 - prob`
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[0, 1], [1, 0]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is the mixed state to |0>
    expected_fidelity = 1 - prob
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_dephasing_state_fidelity(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # prepare noisy dephasing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for dephasing channel
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is the mixed state to |0>
    expected_fidelity = 1
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_depolarizing_state_fidelity(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # prepare noisy depolarizing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for depolarizing channel
    kraus_ops = [
        np.sqrt(3 * prob + 1) / 2 * np.array([[1, 0], [0, 1]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, 1], [1, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, -1j], [1j, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is the mixed state to |0>
    expected_fidelity = (1 + prob) / 2
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_unitary_state_fidelity(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # rotate |0> state by some random angle about X axis
    theta = np.random.uniform(0.0, 2 * np.pi)
    p = Program(RX(theta, 0))

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is this state to |0>
    expected_fidelity = (np.cos(theta / 2)) ** 2
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_bit_flip_state_fidelity_readout_error(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # prepare noisy bit-flip channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # the bit flip channel is composed of two Kraus operations --
    # applying the X gate with probability `prob`, and applying the identity gate
    # with probability `1 - prob`
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[0, 1], [1, 0]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)
    p.define_noisy_readout(0, 0.95, 0.76)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is the mixed state to |0>
    expected_fidelity = 1 - prob
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_dephasing_state_fidelity_readout_error(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # prepare noisy dephasing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for dephasing channel
    kraus_ops = [
        np.sqrt(1 - prob) * np.array([[1, 0], [0, 1]]),
        np.sqrt(prob) * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)
    p.define_noisy_readout(0, 0.95, 0.76)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is the mixed state to |0>
    expected_fidelity = 1
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_depolarizing_state_fidelity_readout_error(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # prepare noisy depolarizing channel as program for some random value of probability
    prob = np.random.uniform(0.1, 0.5)
    # Kraus operators for depolarizing channel
    kraus_ops = [
        np.sqrt(3 * prob + 1) / 2 * np.array([[1, 0], [0, 1]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, 1], [1, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[0, -1j], [1j, 0]]),
        np.sqrt(1 - prob) / 2 * np.array([[1, 0], [0, -1]]),
    ]
    p = Program(Pragma("PRESERVE_BLOCK"), I(0), Pragma("END_PRESERVE_BLOCK"))
    p.define_noisy_gate("I", [0], kraus_ops)
    p.define_noisy_readout(0, 0.95, 0.76)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is the mixed state to |0>
    expected_fidelity = (1 + prob) / 2
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)


def test_unitary_state_fidelity_readout_error(client: Client, use_seed: bool):
    qc = get_qc("1q-qvm", client=client)
    if use_seed:
        qc.qam.random_seed = 0
        np.random.seed(0)
        num_expts = 1
    else:
        num_expts = 100
    # prepare experiment setting
    expt = ExperimentSetting(TensorProductState(), sZ(0))

    # rotate |0> state by some random angle about X axis
    theta = np.random.uniform(0.0, 2 * np.pi)
    p = Program(RX(theta, 0))
    p.define_noisy_readout(0, 0.95, 0.76)

    # prepare TomographyExperiment
    process_exp = Experiment(settings=[expt], program=p)
    # list to store experiment results
    expts = []
    for _ in range(num_expts):
        expt_results = []
        for res in measure_observables(qc, process_exp, n_shots=2000):
            expt_results.append(res.expectation)
        expts.append(expt_results)

    expts = np.array(expts)
    results = np.mean(expts, axis=0)
    estimated_fidelity = _point_state_fidelity_estimate(results)
    # how close is this state to |0>
    expected_fidelity = (np.cos(theta / 2)) ** 2
    np.testing.assert_allclose(expected_fidelity, estimated_fidelity, atol=2e-2)



from math import pi

import numpy as np

from pyquil.quilatom import (
    Parameter,
    quil_sin,
    quil_cos,
    quil_sqrt,
    quil_exp,
    _contained_parameters,
    format_parameter,
    quil_cis,
    substitute,
    substitute_array,
)


def test_format_parameter():
    test_cases = [
        (1, "1"),
        (1.0, "1.0"),
        (1j, "i"),
        (0 + 1j, "i"),
        (-1j, "-i"),
        (1e-15 + 1j, "i"),
        (1e-15 - 1j, "-i"),
    ]

    for test_case in test_cases:
        assert format_parameter(test_case[0]) == test_case[1]


# https://github.com/rigetti/pyquil/issues/184
def test_pretty_print_pi():
    test_cases = [
        (0.0, "0"),
        (pi, "pi"),
        (-pi, "-pi"),
        (2 * pi / 3.0, "2*pi/3"),
        (pi / 9, "0.3490658503988659"),
        (pi / 8, "pi/8"),
        (-90 * pi / 2, "-45*pi"),
    ]

    for test_case in test_cases:
        assert format_parameter(test_case[0]) == test_case[1]


def test_expression_to_string():
    x = Parameter("x")
    assert str(x) == "%x"

    y = Parameter("y")
    assert str(y) == "%y"

    assert str(x + y) == "%x + %y"
    assert str(3 * x + y) == "3*%x + %y"
    assert str(3 * (x + y)) == "3*(%x + %y)"

    assert str(x + y + 2) == "%x + %y + 2"
    assert str(x - y - 2) == "%x - %y - 2"
    assert str(x - (y - 2)) == "%x - (%y - 2)"

    assert str((x + y) - 2) == "%x + %y - 2"
    assert str(x + (y - 2)) == "%x + %y - 2"

    assert str(x ** y ** 2) == "%x^%y^2"
    assert str(x ** (y ** 2)) == "%x^%y^2"
    assert str((x ** y) ** 2) == "(%x^%y)^2"

    assert str(quil_sin(x)) == "SIN(%x)"
    assert str(3 * quil_sin(x + y)) == "3*SIN(%x + %y)"


def test_contained_parameters():
    x = Parameter("x")
    assert _contained_parameters(x) == {x}

    y = Parameter("y")
    assert _contained_parameters(x + y) == {x, y}

    assert _contained_parameters(x ** y ** quil_sin(x * y * 4)) == {x, y}


def test_eval():
    x = Parameter("x")
    assert substitute(x, {x: 5}) == 5

    y = Parameter("y")
    assert substitute(x + y, {x: 5, y: 6}) == 11
    assert substitute(x + y, {x: 5}) == 5 + y
    assert substitute(quil_exp(x), {y: 5}) != np.exp(5)
    assert substitute(quil_exp(x), {x: 5}) == np.exp(5)

    assert np.isclose(substitute(quil_sin(x * x ** 2 / y), {x: 5.0, y: 10.0}), np.sin(12.5))
    assert np.isclose(substitute(quil_sqrt(x), {x: 5.0, y: 10.0}), np.sqrt(5.0))
    assert np.isclose(substitute(quil_cis(x), {x: 5.0, y: 10.0}), np.exp(1j * 5.0))
    assert np.isclose(substitute(x - y, {x: 5.0, y: 10.0}), -5.0)

    assert substitute(quil_cis(x), {y: 5}) == quil_cis(x)
    assert np.allclose(substitute_array([quil_sin(x), quil_cos(x)], {x: 5}), [np.sin(5), np.cos(5)])



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for creating and defining Quil programs.
"""
import itertools
import types
import warnings
from collections import defaultdict
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
    no_type_check,
)

import numpy as np
from rpcq.messages import NativeQuilMetadata

from pyquil._parser.parser import run_parser
from pyquil.gates import MEASURE, RESET
from pyquil.noise import _check_kraus_ops, _create_kraus_pragmas, pauli_kraus_map
from pyquil.quilatom import (
    Label,
    LabelPlaceholder,
    MemoryReference,
    MemoryReferenceDesignator,
    Parameter,
    ParameterDesignator,
    Frame,
    Qubit,
    QubitDesignator,
    QubitPlaceholder,
    FormalArgument,
    format_parameter,
    unpack_classical_reg,
    unpack_qubit,
)
from pyquil.quilbase import (
    DefGate,
    Gate,
    Measurement,
    Pragma,
    Halt,
    AbstractInstruction,
    Jump,
    JumpConditional,
    JumpTarget,
    JumpUnless,
    JumpWhen,
    Declare,
    Reset,
    ResetQubit,
    DelayFrames,
    DelayQubits,
    Fence,
    FenceAll,
    Pulse,
    Capture,
    RawCapture,
    SetFrequency,
    ShiftFrequency,
    SetPhase,
    ShiftPhase,
    SwapPhase,
    SetScale,
    DefPermutationGate,
    DefCalibration,
    DefFrame,
    DefMeasureCalibration,
    DefWaveform,
)
from pyquil.quiltcalibrations import (
    CalibrationError,
    CalibrationMatch,
    expand_calibration,
    match_calibration,
)

InstructionDesignator = Union[
    AbstractInstruction,
    DefGate,
    "Program",
    List[Any],
    Tuple[Any, ...],
    str,  # required to be a pyquil program
    Generator[Any, Any, Any],
]


class Program(object):
    """A list of pyQuil instructions that comprise a quantum program.

    >>> from pyquil import Program
    >>> from pyquil.gates import H, CNOT
    >>> p = Program()
    >>> p += H(0)
    >>> p += CNOT(0, 1)
    """

    def __init__(self, *instructions: InstructionDesignator):
        self._defined_gates: List[DefGate] = []

        self._calibrations: List[Union[DefCalibration, DefMeasureCalibration]] = []
        self._waveforms: Dict[str, DefWaveform] = {}
        self._frames: Dict[Frame, DefFrame] = {}

        # Implementation note: the key difference between the private _instructions and
        # the public instructions property below is that the private _instructions list
        # may contain placeholder labels.
        self._instructions: List[AbstractInstruction] = []

        # Performance optimization: as stated above _instructions may contain placeholder
        # labels so the program must first be have its labels instantiated.
        # _synthesized_instructions is simply a cache on the result of the _synthesize()
        # method.  It is marked as None whenever new instructions are added.
        self._synthesized_instructions: Optional[List[AbstractInstruction]] = None

        # "ro" is always implicitly declared
        self._declarations: Dict[str, Declare] = {"ro": Declare("ro", "BIT")}

        self.inst(*instructions)

        # Filled in with quil_to_native_quil
        self.native_quil_metadata: Optional[NativeQuilMetadata] = None

        # default number of shots to loop through
        self.num_shots = 1

        # Note to developers: Have you changed this method? Have you changed the fields which
        # live on `Program`? Please update `Program.copy()`!

    @property
    def calibrations(self) -> List[Union[DefCalibration, DefMeasureCalibration]]:
        """ A list of Quil-T calibration definitions. """
        return self._calibrations

    @property
    def waveforms(self) -> Dict[str, DefWaveform]:
        """ A mapping from waveform names to their corresponding definitions. """
        return self._waveforms

    @property
    def frames(self) -> Dict[Frame, DefFrame]:
        """ A mapping from Quil-T frames to their definitions. """
        return self._frames

    @property
    def declarations(self) -> Dict[str, Declare]:
        """ A mapping from declared region names to their declarations. """
        return self._declarations

    def copy_everything_except_instructions(self) -> "Program":
        """
        Copy all the members that live on a Program object.

        :return: a new Program
        """
        new_prog = Program()
        new_prog._calibrations = self.calibrations.copy()
        new_prog._waveforms = self.waveforms.copy()
        new_prog._defined_gates = self._defined_gates.copy()
        new_prog._frames = self.frames.copy()
        if self.native_quil_metadata is not None:
            # TODO: remove this type: ignore once rpcq._base.Message gets type hints.
            new_prog.native_quil_metadata = self.native_quil_metadata.copy()  # type: ignore
        new_prog.num_shots = self.num_shots
        return new_prog

    def copy(self) -> "Program":
        """
        Perform a shallow copy of this program.

        QuilAtom and AbstractInstruction objects should be treated as immutable to avoid
        strange behavior when performing a copy.

        :return: a new Program
        """
        new_prog = self.copy_everything_except_instructions()
        new_prog._instructions = self._instructions.copy()
        return new_prog

    @property
    def defined_gates(self) -> List[DefGate]:
        """
        A list of defined gates on the program.
        """
        return self._defined_gates

    @property
    def instructions(self) -> List[AbstractInstruction]:
        """
        Fill in any placeholders and return a list of quil AbstractInstructions.
        """
        if self._synthesized_instructions is None:
            self._synthesize()
        assert self._synthesized_instructions is not None
        return self._synthesized_instructions

    def inst(self, *instructions: InstructionDesignator) -> "Program":
        """
        Mutates the Program object by appending new instructions.

        This function accepts a number of different valid forms, e.g.

            >>> p = Program()
            >>> p.inst(H(0)) # A single instruction
            >>> p.inst(H(0), H(1)) # Multiple instructions
            >>> p.inst([H(0), H(1)]) # A list of instructions
            >>> p.inst(H(i) for i in range(4)) # A generator of instructions
            >>> p.inst(("H", 1)) # A tuple representing an instruction
            >>> p.inst("H 0") # A string representing an instruction
            >>> q = Program()
            >>> p.inst(q) # Another program

        It can also be chained:
            >>> p = Program()
            >>> p.inst(H(0)).inst(H(1))

        :param instructions: A list of Instruction objects, e.g. Gates
        :return: self for method chaining
        """
        for instruction in instructions:
            if isinstance(instruction, list):
                self.inst(*instruction)
            elif isinstance(instruction, types.GeneratorType):
                self.inst(*instruction)
            elif isinstance(instruction, tuple):
                if len(instruction) == 0:
                    raise ValueError("tuple should have at least one element")
                elif len(instruction) == 1:
                    self.inst(instruction[0])
                else:
                    op = instruction[0]
                    if op == "MEASURE":
                        if len(instruction) == 2:
                            self.measure(instruction[1], None)
                        else:
                            self.measure(instruction[1], instruction[2])
                    else:
                        params: List[ParameterDesignator] = []
                        possible_params = instruction[1]
                        rest: Sequence[Any] = instruction[2:]
                        if isinstance(possible_params, list):
                            params = possible_params
                        else:
                            rest = [possible_params] + list(rest)
                        self.gate(op, params, rest)
            elif isinstance(instruction, str):
                self.inst(run_parser(instruction.strip()))
            elif isinstance(instruction, Program):
                if id(self) == id(instruction):
                    raise ValueError("Nesting a program inside itself is not supported")

                for defgate in instruction._defined_gates:
                    self.inst(defgate)
                for instr in instruction._instructions:
                    self.inst(instr)

            # Implementation note: these two base cases are the only ones which modify the program
            elif isinstance(instruction, DefGate):
                defined_gate_names = [gate.name for gate in self._defined_gates]
                if instruction.name in defined_gate_names:
                    warnings.warn("Gate {} has already been defined in this program".format(instruction.name))

                self._defined_gates.append(instruction)
            elif isinstance(instruction, DefCalibration) or isinstance(instruction, DefMeasureCalibration):
                self.calibrations.append(instruction)
            elif isinstance(instruction, DefWaveform):
                self.waveforms[instruction.name] = instruction
            elif isinstance(instruction, DefFrame):
                self.frames[instruction.frame] = instruction
            elif isinstance(instruction, AbstractInstruction):
                self._instructions.append(instruction)
                self._synthesized_instructions = None

                if isinstance(instruction, Declare):
                    self._declarations[instruction.name] = instruction
            else:
                raise TypeError("Invalid instruction: {}".format(instruction))

        return self

    def gate(
        self,
        name: str,
        params: Iterable[ParameterDesignator],
        qubits: Iterable[Union[Qubit, QubitPlaceholder]],
    ) -> "Program":
        """
        Add a gate to the program.

        .. note::

            The matrix elements along each axis are ordered by bitstring. For two qubits the order
            is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
            i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
            See also :ref:`the related docs in the WavefunctionSimulator Overview <basis_ordering>`.

        :param name: The name of the gate.
        :param params: Parameters to send to the gate.
        :param qubits: Qubits that the gate operates on.
        :return: The Program instance
        """
        return self.inst(Gate(name, params, [unpack_qubit(q) for q in qubits]))

    def defgate(
        self,
        name: str,
        matrix: Union[List[List[Any]], np.ndarray, np.matrix],
        parameters: Optional[List[Parameter]] = None,
    ) -> "Program":
        """
        Define a new static gate.

        .. note::

            The matrix elements along each axis are ordered by bitstring. For two qubits the order
            is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
            i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
            See also :ref:`the related docs in the WavefunctionSimulator Overview <basis_ordering>`.


        :param name: The name of the gate.
        :param matrix: List of lists or Numpy 2d array.
        :param parameters: list of parameters that are used in this gate
        :return: The Program instance.
        """
        return self.inst(DefGate(name, matrix, parameters))

    def define_noisy_gate(self, name: str, qubit_indices: Sequence[int], kraus_ops: Sequence[Any]) -> "Program":
        """
        Overload a static ideal gate with a noisy one defined in terms of a Kraus map.

        .. note::

            The matrix elements along each axis are ordered by bitstring. For two qubits the order
            is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
            i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
            See also :ref:`the related docs in the WavefunctionSimulator Overview <basis_ordering>`.


        :param name: The name of the gate.
        :param qubit_indices: The qubits it acts on.
        :param kraus_ops: The Kraus operators.
        :return: The Program instance
        """
        kraus_ops = [np.asarray(k, dtype=np.complex128) for k in kraus_ops]
        _check_kraus_ops(len(qubit_indices), kraus_ops)
        return self.inst(_create_kraus_pragmas(name, tuple(qubit_indices), kraus_ops))

    def define_noisy_readout(self, qubit: Union[int, QubitPlaceholder], p00: float, p11: float) -> "Program":
        """
        For this program define a classical bit flip readout error channel parametrized by
        ``p00`` and ``p11``. This models the effect of thermal noise that corrupts the readout
        signal **after** it has interrogated the qubit.

        :param qubit: The qubit with noisy readout.
        :param p00: The probability of obtaining the measurement result 0 given that the qubit
          is in state 0.
        :param p11: The probability of obtaining the measurement result 1 given that the qubit
          is in state 1.
        :return: The Program with an appended READOUT-POVM Pragma.
        """
        if not 0.0 <= p00 <= 1.0:
            raise ValueError("p00 must be in the interval [0,1].")
        if not 0.0 <= p11 <= 1.0:
            raise ValueError("p11 must be in the interval [0,1].")
        if not (isinstance(qubit, int) or isinstance(qubit, QubitPlaceholder)):
            raise TypeError("qubit must be a non-negative integer, or QubitPlaceholder.")
        if isinstance(qubit, int) and qubit < 0:
            raise ValueError("qubit cannot be negative.")
        p00 = float(p00)
        p11 = float(p11)
        aprobs = [p00, 1.0 - p11, 1.0 - p00, p11]
        aprobs_str = "({})".format(" ".join(format_parameter(p) for p in aprobs))
        pragma = Pragma("READOUT-POVM", [qubit], aprobs_str)
        return self.inst(pragma)

    def no_noise(self) -> "Program":
        """
        Prevent a noisy gate definition from being applied to the immediately following Gate
        instruction.

        :return: Program
        """
        return self.inst(Pragma("NO-NOISE"))

    def measure(self, qubit: QubitDesignator, classical_reg: Optional[MemoryReferenceDesignator]) -> "Program":
        """
        Measures a qubit at qubit_index and puts the result in classical_reg

        :param qubit: The qubit to measure.
        :param classical_reg: The classical register to measure into, or None.
        :returns: The Quil Program with the appropriate measure instruction appended, e.g.
                  MEASURE 0 [1]
        """
        return self.inst(MEASURE(qubit, classical_reg))

    def reset(self, qubit_index: Optional[int] = None) -> "Program":
        """
        Reset all qubits or just a specific qubit at qubit_index.

        :param qubit_index: The address of the qubit to reset.
            If None, reset all qubits.
        :returns: The Quil Program with the appropriate reset instruction appended, e.g.
                  RESET 0
        """
        return self.inst(RESET(qubit_index))

    def measure_all(self, *qubit_reg_pairs: Tuple[QubitDesignator, Optional[MemoryReferenceDesignator]]) -> "Program":
        """
        Measures many qubits into their specified classical bits, in the order
        they were entered. If no qubit/register pairs are provided, measure all qubits present in
        the program into classical addresses of the same index.

        :param qubit_reg_pairs: Tuples of qubit indices paired with classical bits.
        :return: The Quil Program with the appropriate measure instructions appended, e.g.

        .. code::

                  MEASURE 0 [1]
                  MEASURE 1 [2]
                  MEASURE 2 [3]
        """
        if qubit_reg_pairs == ():
            qubits = self.get_qubits(indices=True)
            if len(qubits) == 0:
                return self
            if any(isinstance(q, QubitPlaceholder) for q in qubits):
                raise ValueError(
                    "Attempted to call measure_all on a Program that contains QubitPlaceholders. "
                    "You must either provide the qubit_reg_pairs argument to describe how to map "
                    "these QubitPlaceholders to memory registers, or else first call "
                    "pyquil.quil.address_qubits to instantiate the QubitPlaceholders."
                )
            if any(isinstance(q, FormalArgument) for q in qubits):
                raise ValueError("Cannot call measure_all on a Program that contains FormalArguments.")
            # Help mypy determine that qubits does not contain any QubitPlaceholders.
            qubit_inds = cast(List[int], qubits)
            ro = self.declare("ro", "BIT", max(qubit_inds) + 1)
            for qi in qubit_inds:
                self.inst(MEASURE(qi, ro[qi]))
        else:
            for qubit_index, classical_reg in qubit_reg_pairs:
                self.inst(MEASURE(qubit_index, classical_reg))
        return self

    def while_do(self, classical_reg: MemoryReferenceDesignator, q_program: "Program") -> "Program":
        """
        While a classical register at index classical_reg is 1, loop q_program

        Equivalent to the following construction:

        .. code::

            WHILE [c]:
               instr...
            =>
              LABEL @START
              JUMP-UNLESS @END [c]
              instr...
              JUMP @START
              LABEL @END

        :param MemoryReferenceDesignator classical_reg: The classical register to check
        :param Program q_program: The Quil program to loop.
        :return: The Quil Program with the loop instructions added.
        """
        label_start = LabelPlaceholder("START")
        label_end = LabelPlaceholder("END")
        self.inst(JumpTarget(label_start))
        self.inst(JumpUnless(target=label_end, condition=unpack_classical_reg(classical_reg)))
        self.inst(q_program)
        self.inst(Jump(label_start))
        self.inst(JumpTarget(label_end))
        return self

    def if_then(
        self,
        classical_reg: MemoryReferenceDesignator,
        if_program: "Program",
        else_program: Optional["Program"] = None,
    ) -> "Program":
        """
        If the classical register at index classical reg is 1, run if_program, else run
        else_program.

        Equivalent to the following construction:

        .. code::

            IF [c]:
               instrA...
            ELSE:
               instrB...
            =>
              JUMP-WHEN @THEN [c]
              instrB...
              JUMP @END
              LABEL @THEN
              instrA...
              LABEL @END

        :param classical_reg: The classical register to check as the condition
        :param if_program: A Quil program to execute if classical_reg is 1
        :param else_program: A Quil program to execute if classical_reg is 0. This
            argument is optional and defaults to an empty Program.
        :returns: The Quil Program with the branching instructions added.
        """
        else_program = else_program if else_program is not None else Program()

        label_then = LabelPlaceholder("THEN")
        label_end = LabelPlaceholder("END")
        self.inst(JumpWhen(target=label_then, condition=unpack_classical_reg(classical_reg)))
        self.inst(else_program)
        self.inst(Jump(label_end))
        self.inst(JumpTarget(label_then))
        self.inst(if_program)
        self.inst(JumpTarget(label_end))
        return self

    def alloc(self) -> QubitPlaceholder:
        """
        Get a new qubit.

        :return: A qubit.
        """
        warnings.warn(
            "`alloc` is deprecated and will be removed in a future version of pyQuil. "
            "Please create a `QubitPlaceholder` directly",
            DeprecationWarning,
        )
        return QubitPlaceholder()

    def declare(
        self,
        name: str,
        memory_type: str = "BIT",
        memory_size: int = 1,
        shared_region: Optional[str] = None,
        offsets: Optional[Iterable[Tuple[int, str]]] = None,
    ) -> MemoryReference:
        """DECLARE a quil variable

        This adds the declaration to the current program and returns a MemoryReference to the
        base (offset = 0) of the declared memory.

        .. note::
            This function returns a MemoryReference and cannot be chained like some
            of the other Program methods. Consider using ``inst(DECLARE(...))`` if you
            would like to chain methods, but please be aware that you must create your
            own MemoryReferences later on.

        :param name: Name of the declared variable
        :param memory_type: Type of the declared memory: 'BIT', 'REAL', 'OCTET' or 'INTEGER'
        :param memory_size: Number of array elements in the declared memory.
        :param shared_region: You can declare a variable that shares its underlying memory
            with another region. This allows aliasing. For example, you can interpret an array
            of measured bits as an integer.
        :param offsets: If you are using ``shared_region``, this allows you to share only
            a part of the parent region. The offset is given by an array type and the number
            of elements of that type. For example,
            ``DECLARE target-bit BIT SHARING real-region OFFSET 1 REAL 4 BIT`` will let you use
            target-bit to poke into the fourth bit of the second real from the leading edge of
            real-region.
        :return: a MemoryReference to the start of the declared memory region, ie a memory
            reference to ``name[0]``.
        """
        self.inst(
            Declare(
                name=name,
                memory_type=memory_type,
                memory_size=memory_size,
                shared_region=shared_region,
                offsets=offsets,
            )
        )
        return MemoryReference(name=name, declared_size=memory_size)

    def wrap_in_numshots_loop(self, shots: int) -> "Program":
        """
        Wraps a Quil program in a loop that re-runs the same program many times.

        Note: this function is a prototype of what will exist in the future when users will
        be responsible for writing this loop instead of having it happen automatically.

        :param shots: Number of iterations to loop through.
        """
        self.num_shots = shots
        return self

    def out(self, *, calibrations: Optional[bool] = True) -> str:
        """
        Serializes the Quil program to a string suitable for submitting to the QVM or QPU.
        """

        return "\n".join(
            itertools.chain(
                (dg.out() for dg in self._defined_gates),
                (wf.out() for wf in self.waveforms.values()),
                (fdef.out() for fdef in self.frames.values()),
                (cal.out() for cal in self.calibrations) if calibrations else list(),
                (instr.out() for instr in self.instructions),
                [""],
            )
        )

    def get_qubits(self, indices: bool = True) -> Set[QubitDesignator]:
        """
        Returns all of the qubit indices used in this program, including gate applications and
        allocated qubits. e.g.

            >>> p = Program()
            >>> p.inst(("H", 1))
            >>> p.get_qubits()
            {1}
            >>> q = p.alloc()
            >>> p.inst(H(q))
            >>> len(p.get_qubits())
            2

        :param indices: Return qubit indices as integers intead of the
            wrapping :py:class:`Qubit` object
        :return: A set of all the qubit indices used in this program
        """
        qubits: Set[QubitDesignator] = set()
        for instr in self.instructions:
            if isinstance(
                instr,
                (
                    Gate,
                    Measurement,
                    ResetQubit,
                    Pulse,
                    Capture,
                    RawCapture,
                    ShiftFrequency,
                    SetFrequency,
                    SetPhase,
                    ShiftPhase,
                    SwapPhase,
                    SetScale,
                ),
            ):
                qubits |= instr.get_qubits(indices=indices)
        return qubits

    def match_calibrations(self, instr: AbstractInstruction) -> Optional[CalibrationMatch]:
        """
        Attempt to match a calibration to the provided instruction.

        Note: preference is given to later calibrations, i.e. in a program with

          DEFCAL X 0:
              <a>

          DEFCAL X 0:
             <b>

        the second calibration, with body <b>, would be the calibration matching `X 0`.

        :param instr: An instruction.
        :returns: a CalibrationMatch object, if one can be found.
        """
        if isinstance(instr, (Gate, Measurement)):
            for cal in reversed(self.calibrations):
                match = match_calibration(instr, cal)
                if match is not None:
                    return match

        return None

    def get_calibration(self, instr: AbstractInstruction) -> Optional[Union[DefCalibration, DefMeasureCalibration]]:
        """
        Get the calibration corresponding to the provided instruction.

        :param instr: An instruction.
        :returns: A matching Quil-T calibration definition, if one exists.
        """
        match = self.match_calibrations(instr)
        if match:
            return match.cal

        return None

    def calibrate(
        self,
        instruction: AbstractInstruction,
        previously_calibrated_instructions: Optional[Set[AbstractInstruction]] = None,
    ) -> List[AbstractInstruction]:
        """
        Expand an instruction into its calibrated definition.

        If a calibration definition matches the provided instruction, then the definition
        body is returned with appropriate substitutions made for parameters and qubit
        arguments. If no calibration definition matches, then the original instruction
        is returned. Calibrations are performed recursively, so that if a calibrated
        instruction produces an instruction that has a corresponding calibration, it
        will be expanded, and so on. If a cycle is encountered, a CalibrationError is
        raised.

        :param instruction: An instruction.
        :param previously_calibrated_instructions: A set of instructions that are the
            results of calibration expansions in the direct ancestry of `instruction`.
            Used to catch cyclic calibration expansions.
        :returns: A list of instructions, with the active calibrations expanded.
        """
        if previously_calibrated_instructions is None:
            previously_calibrated_instructions = set()
        elif instruction in previously_calibrated_instructions:
            raise CalibrationError(
                f"The instruction {instruction} appears in the set of "
                f"previously calibrated instructions {previously_calibrated_instructions}"
                " and would therefore result in a cyclic non-terminating expansion."
            )
        else:
            previously_calibrated_instructions = previously_calibrated_instructions.union({instruction})
        match = self.match_calibrations(instruction)
        if match is not None:
            return sum(
                [
                    self.calibrate(expansion, previously_calibrated_instructions)
                    for expansion in expand_calibration(match)
                ],
                [],
            )
        else:
            return [instruction]

    def is_protoquil(self, quilt: bool = False) -> bool:
        """
        Protoquil programs may only contain gates, Pragmas, and RESET. It may not contain
        classical instructions or jumps.

        :return: True if the Program is Protoquil, False otherwise
        """
        try:
            if quilt:
                validate_protoquil(self, quilt=quilt)
            else:
                validate_protoquil(self)
            return True
        except ValueError:
            return False

    def is_supported_on_qpu(self) -> bool:
        """
        Whether the program can be compiled to the hardware to execute on a QPU. These Quil
        programs are more restricted than Protoquil: for instance, RESET must be before any
        gates or MEASUREs, and MEASURE on a qubit must be after any gates on that qubit.

        :return: True if the Program is supported Quil, False otherwise
        """
        try:
            validate_supported_quil(self)
            return True
        except ValueError:
            return False

    def pop(self) -> AbstractInstruction:
        """
        Pops off the last instruction.

        :return: The instruction that was popped.
        """
        res = self._instructions.pop()
        self._synthesized_instructions = None
        return res

    def dagger(self, inv_dict: Optional[Any] = None, suffix: str = "-INV") -> "Program":
        """
        Creates the conjugate transpose of the Quil program. The program must
        contain only gate applications.

        Note: the keyword arguments inv_dict and suffix are kept only
        for backwards compatibility and have no effect.

        :return: The Quil program's inverse
        """
        if any(not isinstance(instr, Gate) for instr in self._instructions):
            raise ValueError("Program to be daggered must contain only gate applications")

        # This is a bit hacky. Gate.dagger() mutates the gate object, rather than returning a fresh
        # (and daggered) copy. Also, mypy doesn't understand that we already asserted that every
        # instr in _instructions is a Gate, above, so help mypy out with a cast.
        surely_gate_instructions = cast(List[Gate], Program(self.out())._instructions)
        return Program([instr.dagger() for instr in reversed(surely_gate_instructions)])

    def _synthesize(self) -> "Program":
        """
        Assigns all placeholder labels to actual values and implicitly declares the ``ro``
        register for backwards compatibility.

        Changed in 1.9: Either all qubits must be defined or all undefined. If qubits are
        undefined, this method will not help you. You must explicitly call `address_qubits`
        which will return a new Program.

        Changed in 1.9: This function now returns ``self`` and updates
        ``self._synthesized_instructions``.

        Changed in 2.0: This function will add an instruction to the top of the program
        to declare a register of bits called ``ro`` if and only if there are no other
        declarations in the program.

        :return: This object with the ``_synthesized_instructions`` member set.
        """
        self._synthesized_instructions = instantiate_labels(self._instructions)
        self._synthesized_instructions = implicitly_declare_ro(self._synthesized_instructions)
        return self

    def __add__(self, other: InstructionDesignator) -> "Program":
        """
        Concatenate two programs together, returning a new one.

        :param other: Another program or instruction to concatenate to this one.
        :return: A newly concatenated program.
        """
        p = Program()
        p.inst(self)
        p.inst(other)
        p._calibrations = self.calibrations
        p._waveforms = self.waveforms
        p._frames = self.frames
        if isinstance(other, Program):
            p.calibrations.extend(other.calibrations)
            p.waveforms.update(other.waveforms)
            p.frames.update(other.frames)
        return p

    def __iadd__(self, other: InstructionDesignator) -> "Program":
        """
        Concatenate two programs together using +=, returning a new one.

        :param other: Another program or instruction to concatenate to this one.
        :return: A newly concatenated program.
        """
        self.inst(other)
        if isinstance(other, Program):
            self.calibrations.extend(other.calibrations)
            self.waveforms.update(other.waveforms)
            self.frames.update(other.frames)
        return self

    def __getitem__(self, index: Union[slice, int]) -> Union[AbstractInstruction, "Program"]:
        """
        Allows indexing into the program to get an action.

        :param index: The action at the specified index.
        :return:
        """
        return Program(self.instructions[index]) if isinstance(index, slice) else self.instructions[index]

    def __iter__(self) -> Iterator[AbstractInstruction]:
        """
        Allow built in iteration through a program's instructions, e.g. [a for a in Program(X(0))]

        :return:
        """
        return self.instructions.__iter__()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.out() == other.out()

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __len__(self) -> int:
        return len(self.instructions)

    def __str__(self) -> str:
        """
        A string representation of the Quil program for inspection.

        This may not be suitable for submission to a QPU or QVM for example if
        your program contains unaddressed QubitPlaceholders
        """
        return "\n".join(
            itertools.chain(
                (str(dg) for dg in self._defined_gates),
                (str(wf) for wf in self.waveforms.values()),
                (str(fdef) for fdef in self.frames.values()),
                (str(cal) for cal in self.calibrations),
                (str(instr) for instr in self.instructions),
                [""],
            )
        )


def _what_type_of_qubit_does_it_use(
    program: Program,
) -> Tuple[bool, bool, List[Union[Qubit, QubitPlaceholder]]]:
    """Helper function to peruse through a program's qubits.

    This function will also enforce the condition that a Program uses either all placeholders
    or all instantiated qubits to avoid accidentally mixing the two. This function will warn
    if your program doesn't use any qubits.

    :return: tuple of (whether the program uses placeholder qubits, whether the program uses
        real qubits, a list of qubits ordered by their first appearance in the program)
    """
    has_placeholders = False
    has_real_qubits = False

    # We probably want to index qubits in the order they are encountered in the program
    # so an ordered set would be nice. Python doesn't *have* an ordered set. Use the keys
    # of an ordered dictionary instead
    qubits = {}

    for instr in program:
        if isinstance(instr, Gate):
            for q in instr.qubits:
                qubits[q] = 1
                if isinstance(q, QubitPlaceholder):
                    has_placeholders = True
                elif isinstance(q, Qubit):
                    has_real_qubits = True
                else:
                    raise ValueError("Unknown qubit type {}".format(q))
        elif isinstance(instr, Measurement):
            qubits[instr.qubit] = 1
            if isinstance(instr.qubit, QubitPlaceholder):
                has_placeholders = True
            elif isinstance(instr.qubit, Qubit):
                has_real_qubits = True
            else:
                raise ValueError("Unknown qubit type {}".format(instr.qubit))
        elif isinstance(instr, Pragma):
            for arg in instr.args:
                if isinstance(arg, QubitPlaceholder):
                    qubits[arg] = 1
                    has_placeholders = True
                elif isinstance(arg, Qubit):
                    qubits[arg] = 1
                    has_real_qubits = True
    if not (has_placeholders or has_real_qubits):
        warnings.warn("Your program doesn't use any qubits")

    if has_placeholders and has_real_qubits:
        raise ValueError("Your program mixes instantiated qubits with placeholders")

    # The isinstance checks above make sure that if any qubit is a
    # FormalArgument (which is permitted by Gate.qubits), then an
    # error should be raised. Unfortunately this doesn't help mypy
    # narrow down the return type, so gotta cast.
    return (
        has_placeholders,
        has_real_qubits,
        cast(List[Union[Qubit, QubitPlaceholder]], list(qubits.keys())),
    )


def get_default_qubit_mapping(program: Program) -> Dict[Union[Qubit, QubitPlaceholder], Qubit]:
    """
    Takes a program which contains qubit placeholders and provides a mapping to the integers
    0 through N-1.

    The output of this function is suitable for input to :py:func:`address_qubits`.

    :param program: A program containing qubit placeholders
    :return: A dictionary mapping qubit placeholder to an addressed qubit from 0 through N-1.
    """
    fake_qubits, real_qubits, qubits = _what_type_of_qubit_does_it_use(program)
    if real_qubits:
        warnings.warn("This program contains integer qubits, so getting a mapping doesn't make sense.")
        # _what_type_of_qubit_does_it_use ensures that if real_qubits is True, then qubits contains
        # only real Qubits, not QubitPlaceholders. Help mypy figure this out with cast.
        return {q: cast(Qubit, q) for q in qubits}
    return {qp: Qubit(i) for i, qp in enumerate(qubits)}


@no_type_check
def address_qubits(
    program: Program, qubit_mapping: Optional[Dict[QubitPlaceholder, Union[Qubit, int]]] = None
) -> Program:
    """
    Takes a program which contains placeholders and assigns them all defined values.

    Either all qubits must be defined or all undefined. If qubits are
    undefined, you may provide a qubit mapping to specify how placeholders get mapped
    to actual qubits. If a mapping is not provided, integers 0 through N are used.

    This function will also instantiate any label placeholders.

    :param program: The program.
    :param qubit_mapping: A dictionary-like object that maps from :py:class:`QubitPlaceholder`
        to :py:class:`Qubit` or ``int`` (but not both).
    :return: A new Program with all qubit and label placeholders assigned to real qubits and labels.
    """
    fake_qubits, real_qubits, qubits = _what_type_of_qubit_does_it_use(program)
    if real_qubits:
        if qubit_mapping is not None:
            warnings.warn("A qubit mapping was provided but the program does not " "contain any placeholders to map!")
        return program

    if qubit_mapping is None:
        qubit_mapping = {qp: Qubit(i) for i, qp in enumerate(qubits)}
    else:
        if all(isinstance(v, Qubit) for v in qubit_mapping.values()):
            pass  # we good
        elif all(isinstance(v, int) for v in qubit_mapping.values()):
            qubit_mapping = {k: Qubit(v) for k, v in qubit_mapping.items()}
        else:
            raise ValueError("Qubit mapping must map to type Qubit or int (but not both)")

    result: List[AbstractInstruction] = []
    for instr in program:
        # Remap qubits on Gate, Measurement, and ResetQubit instructions
        if isinstance(instr, Gate):
            remapped_qubits = [qubit_mapping[q] for q in instr.qubits]
            gate = Gate(instr.name, instr.params, remapped_qubits)
            gate.modifiers = instr.modifiers
            result.append(gate)
        elif isinstance(instr, Measurement):
            result.append(Measurement(qubit_mapping[instr.qubit], instr.classical_reg))
        elif isinstance(instr, ResetQubit):
            result.append(ResetQubit(qubit_mapping[instr.qubit]))
        elif isinstance(instr, Pragma):
            new_args: List[Union[Qubit, int, str]] = []
            for arg in instr.args:
                # Pragmas can have arguments that represent things besides qubits, so here we
                # make sure to just look up the QubitPlaceholders.
                if isinstance(arg, QubitPlaceholder):
                    new_args.append(qubit_mapping[arg])
                else:
                    new_args.append(arg)
            result.append(Pragma(instr.command, new_args, instr.freeform_string))
        # Otherwise simply add it to the result
        else:
            result.append(instr)

    new_program = program.copy()
    new_program._instructions = result

    return new_program


def _get_label(
    placeholder: LabelPlaceholder,
    label_mapping: Dict[LabelPlaceholder, Label],
    label_i: int,
) -> Tuple[Label, Dict[LabelPlaceholder, Label], int]:
    """Helper function to either get the appropriate label for a given placeholder or generate
    a new label and update the mapping.

    See :py:func:`instantiate_labels` for usage.
    """
    if placeholder in label_mapping:
        return label_mapping[placeholder], label_mapping, label_i

    new_target = Label("{}{}".format(placeholder.prefix, label_i))
    label_i += 1
    label_mapping[placeholder] = new_target
    return new_target, label_mapping, label_i


def instantiate_labels(instructions: Iterable[AbstractInstruction]) -> List[AbstractInstruction]:
    """
    Takes an iterable of instructions which may contain label placeholders and assigns
    them all defined values.

    :return: list of instructions with all label placeholders assigned to real labels.
    """
    label_i = 1
    result: List[AbstractInstruction] = []
    label_mapping: Dict[LabelPlaceholder, Label] = dict()
    for instr in instructions:
        if isinstance(instr, Jump) and isinstance(instr.target, LabelPlaceholder):
            new_target, label_mapping, label_i = _get_label(instr.target, label_mapping, label_i)
            result.append(Jump(new_target))
        elif isinstance(instr, JumpConditional) and isinstance(instr.target, LabelPlaceholder):
            new_target, label_mapping, label_i = _get_label(instr.target, label_mapping, label_i)
            cls = instr.__class__  # Make the correct subclass
            result.append(cls(new_target, instr.condition))
        elif isinstance(instr, JumpTarget) and isinstance(instr.label, LabelPlaceholder):
            new_label, label_mapping, label_i = _get_label(instr.label, label_mapping, label_i)
            result.append(JumpTarget(new_label))
        else:
            result.append(instr)

    return result


def implicitly_declare_ro(instructions: List[AbstractInstruction]) -> List[AbstractInstruction]:
    """
    Implicitly declare a register named ``ro`` for backwards compatibility with Quil 1.

    There used to be one un-named hunk of classical memory. Now there are variables with
    declarations. Instead of::

        MEASURE 0 [0]

    You must now measure into a named register, idiomatically::

        MEASURE 0 ro[0]

    The ``MEASURE`` instruction will emit this (with a deprecation warning) if you're still
    using bare integers for classical addresses. However, you must also declare memory in the
    new scheme::

        DECLARE ro BIT[8]
        MEASURE 0 ro[0]

    This method will determine if you are in "backwards compatibility mode" and will declare
    a read-out ``ro`` register for you. If you program contains any DECLARE commands or if it
    does not have any MEASURE x ro[x], this will not do anything.

    This behavior is included for backwards compatibility and will be removed in future releases
    of PyQuil. Please DECLARE all memory including ``ro``.
    """
    ro_addrs: List[int] = []
    for instr in instructions:
        if isinstance(instr, Declare):
            # The user has declared their own memory
            # so they are responsible for all declarations and memory references.
            return instructions

        if isinstance(instr, Measurement):
            if instr.classical_reg is None:
                continue

            if instr.classical_reg.name == "ro":
                ro_addrs += [instr.classical_reg.offset]
            else:
                # The user has used a classical register named something other than "ro"
                # so they are responsible for all declarations and memory references.
                return instructions

    if len(ro_addrs) == 0:
        return instructions

    warnings.warn(
        "Please DECLARE all memory. I'm adding a declaration for the `ro` register, "
        "but I won't do this for you in the future."
    )

    new_instr = instructions.copy()
    new_instr.insert(0, Declare(name="ro", memory_type="BIT", memory_size=max(ro_addrs) + 1))
    return new_instr


def merge_with_pauli_noise(
    prog_list: Iterable[Program], probabilities: Sequence[float], qubits: Sequence[int]
) -> Program:
    """
    Insert pauli noise channels between each item in the list of programs.
    This noise channel is implemented as a single noisy identity gate acting on the provided qubits.
    This method does not rely on merge_programs and so avoids the inclusion of redundant Kraus
    Pragmas that would occur if merge_programs was called directly on programs with distinct noisy
    gate definitions.

    :param prog_list: an iterable such as a program or a list of programs.
        If a program is provided, a single noise gate will be applied after each gate in the
        program. If a list of programs is provided, the noise gate will be applied after each
        program.
    :param probabilities: The 4^num_qubits list of probabilities specifying the desired pauli
        channel. There should be either 4 or 16 probabilities specified in the order
        I, X, Y, Z or II, IX, IY, IZ, XI, XX, XY, etc respectively.
    :param qubits: a list of the qubits that the noisy gate should act on.
    :return: A single program with noisy gates inserted between each element of the program list.
    """
    p = Program()
    p.defgate("pauli_noise", np.eye(2 ** len(qubits)))
    p.define_noisy_gate("pauli_noise", qubits, pauli_kraus_map(probabilities))
    for elem in prog_list:
        p.inst(Program(elem))
        if isinstance(elem, Measurement):
            continue  # do not apply noise after measurement
        p.inst(("pauli_noise", *qubits))
    return p


# TODO: does this need modification?
def merge_programs(prog_list: Sequence[Program]) -> Program:
    """
    Merges a list of pyQuil programs into a single one by appending them in sequence.
    If multiple programs in the list contain the same gate and/or noisy gate definition
    with identical name, this definition will only be applied once. If different definitions
    with the same name appear multiple times in the program list, each will be applied once
    in the order of last occurrence.

    :param prog_list: A list of pyquil programs
    :return: a single pyQuil program
    """
    definitions = [gate for prog in prog_list for gate in Program(prog).defined_gates]
    seen: Dict[str, List[DefGate]] = {}
    # Collect definitions in reverse order and reapply definitions in reverse
    # collected order to ensure that the last occurrence of a definition is applied last.
    for definition in reversed(definitions):
        name = definition.name
        if name in seen.keys():
            # Do not add truly identical definitions with the same name
            # If two different definitions share a name, we include each definition so as to provide
            # a waring to the user when the contradictory defgate is called.
            if definition not in seen[name]:
                seen[name].append(definition)
        else:
            seen[name] = [definition]
    new_definitions = [gate for key in seen.keys() for gate in reversed(seen[key])]

    # Combine programs without gate definitions; avoid call to _synthesize by using _instructions
    p = Program(*[prog._instructions for prog in prog_list])

    for definition in new_definitions:
        if isinstance(definition, DefPermutationGate):
            p.inst(DefPermutationGate(definition.name, list(definition.permutation)))
        else:
            p.defgate(definition.name, definition.matrix, definition.parameters)

    return p


def get_classical_addresses_from_program(program: Program) -> Dict[str, List[int]]:
    """
    Returns a sorted list of classical addresses found in the MEASURE instructions in the program.

    :param program: The program from which to get the classical addresses.
    :return: A mapping from memory region names to lists of offsets appearing in the program.
    """
    addresses: Dict[str, List[int]] = defaultdict(list)
    flattened_addresses = {}

    # Required to use the `classical_reg.address` int attribute.
    # See https://github.com/rigetti/pyquil/issues/388.
    for instr in program:
        if isinstance(instr, Measurement) and instr.classical_reg:
            addresses[instr.classical_reg.name].append(instr.classical_reg.offset)

    # flatten duplicates
    for k, v in addresses.items():
        reduced_list = list(set(v))
        reduced_list.sort()
        flattened_addresses[k] = reduced_list

    return flattened_addresses


def percolate_declares(program: Program) -> Program:
    """
    Move all the DECLARE statements to the top of the program. Return a fresh object.

    :param program: Perhaps jumbled program.
    :return: Program with DECLAREs all at the top and otherwise the same sorted contents.
    """
    declare_program = Program()
    instrs_program = Program()

    for instr in program:
        if isinstance(instr, Declare):
            declare_program += instr
        else:
            instrs_program += instr

    p = declare_program + instrs_program
    p._defined_gates = program._defined_gates

    return p


def validate_protoquil(program: Program, quilt: bool = False) -> None:
    """
    Ensure that a program is valid ProtoQuil or Quil-T, otherwise raise a ValueError.
    Protoquil is a subset of Quil which excludes control flow and classical instructions.

    :param quilt: Validate the program as Quil-T.
    :param program: The Quil program to validate.
    """
    """
    Ensure that a program is valid ProtoQuil, otherwise raise a ValueError.
    Protoquil is a subset of Quil which excludes control flow and classical instructions.

    :param program: The Quil program to validate.
    """
    if quilt:
        valid_instruction_types = tuple(
            [
                Pragma,
                Declare,
                Halt,
                Gate,
                Measurement,
                Reset,
                ResetQubit,
                DelayQubits,
                DelayFrames,
                Fence,
                FenceAll,
                ShiftFrequency,
                SetFrequency,
                SetScale,
                ShiftPhase,
                SetPhase,
                SwapPhase,
                Pulse,
                Capture,
                RawCapture,
                DefCalibration,
                DefFrame,
                DefMeasureCalibration,
                DefWaveform,
            ]
        )
    else:
        valid_instruction_types = tuple([Pragma, Declare, Gate, Reset, ResetQubit, Measurement])
        if program.calibrations:
            raise ValueError("ProtoQuil validation failed: Quil-T calibrations are not allowed.")
        if program.waveforms:
            raise ValueError("ProtoQuil validation failed: Quil-T waveform definitions are not allowed.")
        if program.frames:
            raise ValueError("ProtoQuil validation failed: Quil-T frame definitions are not allowed.")

    for instr in program.instructions:
        if not isinstance(instr, valid_instruction_types):
            # Instructions like MOVE, NOT, JUMP, JUMP-UNLESS will fail here
            raise ValueError(f"ProtoQuil validation failed: {instr} is not allowed.")


def validate_supported_quil(program: Program) -> None:
    """
    Ensure that a program is supported Quil which can run on any QPU, otherwise raise a ValueError.
    We support a global RESET before any gates, and MEASUREs on each qubit after any gates
    on that qubit. PRAGMAs and DECLAREs are always allowed.

    :param program: The Quil program to validate.
    """
    gates_seen = False
    measured_qubits: Set[int] = set()
    for instr in program.instructions:
        if isinstance(instr, Pragma) or isinstance(instr, Declare):
            continue
        elif isinstance(instr, Gate):
            gates_seen = True
            if any(q.index in measured_qubits for q in instr.qubits):
                raise ValueError("Cannot apply gates to qubits that were already measured.")
        elif isinstance(instr, Reset):
            if gates_seen:
                raise ValueError("RESET can only be applied before any gate applications.")
        elif isinstance(instr, ResetQubit):
            raise ValueError("Only global RESETs are currently supported.")
        elif isinstance(instr, Measurement):
            if instr.qubit.index in measured_qubits:
                raise ValueError("Multiple measurements per qubit is not supported.")
            measured_qubits.add(instr.qubit.index)
        else:
            raise ValueError(f"Unhandled instruction type in supported Quil validation: {instr}")



import numpy as np
from pyquil.gates import RZ, RX, I, CZ, ISWAP, CPHASE
from pyquil.noise_gates import _get_qvm_noise_supported_gates, THETA


def test_get_qvm_noise_supported_gates_from_compiler_isa(compiler_isa):
    gates = _get_qvm_noise_supported_gates(compiler_isa)
    for q in [0, 1, 2]:
        for g in [
            I(q),
            RX(np.pi / 2, q),
            RX(-np.pi / 2, q),
            RX(np.pi, q),
            RX(-np.pi, q),
            RZ(THETA, q),
        ]:
            assert g in gates

    assert CZ(0, 1) in gates
    assert CZ(1, 0) in gates
    assert ISWAP(1, 2) in gates
    assert ISWAP(2, 1) in gates
    assert CPHASE(THETA, 2, 0) in gates
    assert CPHASE(THETA, 0, 2) in gates


ASPEN_8_QUBITS_NO_RX = {8, 9, 10, 18, 19, 28, 29, 31}
ASPEN_8_QUBITS_NO_RZ = {8, 9, 10, 18, 19, 28, 29, 31}
ASPEN_8_EDGES_NO_CZ = {(0, 1), (10, 11), (1, 2), (21, 22), (17, 10), (12, 25)}


def test_get_qvm_noise_supported_gates_from_aspen8_isa(qcs_aspen8_device, noise_model_dict):
    gates = _get_qvm_noise_supported_gates(qcs_aspen8_device.to_compiler_isa())

    for q in range(len(qcs_aspen8_device._isa.architecture.nodes)):
        if q not in ASPEN_8_QUBITS_NO_RX:
            for g in [
                RX(np.pi / 2, q),
                RX(-np.pi / 2, q),
                RX(np.pi, q),
                RX(-np.pi, q),
            ]:
                assert g in gates
        if q not in ASPEN_8_QUBITS_NO_RZ:
            assert RZ(THETA, q) in gates

    for edge in qcs_aspen8_device._isa.architecture.edges:
        if (
            edge.node_ids[0],
            edge.node_ids[1],
        ) in ASPEN_8_EDGES_NO_CZ:
            continue
        assert CZ(edge.node_ids[0], edge.node_ids[1]) in gates
        assert CZ(edge.node_ids[1], edge.node_ids[0]) in gates
